import collections
from enum import Enum
from .state import State

import numpy as np


class NodeContext(object):
    """
    The NodeContext consists of the states, identities (as node indices), and adjancency matrix weights of the nodes
    that influence a given node. Each of the properties (i.e. activities, neighbour_indices, connection states) are
    represented as lists. Each of the lists are of the same size, and the first item in each list corresponds to the
    first node in the neighbourhood, etc. The neighbourhood of a node also contains the node's current activity (i.e.
    the activity as of the last timestep).
    """

    __slots__ = ("node_label", "timestep", "activities", "neighbour_labels", "neighbourhood_activities",
                 "connection_states", "current_activity", "past_activities", "input", "added_nodes", "removed_nodes")

    def __init__(self, node_label, timestep, activities, neighbour_labels, neighbourhood_activities, connection_states,
                 current_activity, past_activities, input):
        """
        :param node_label: the label of the node being processed
        :param timestep: the current timestep
        :param activities: the activities of all the nodes after the previous timestep
        :param neighbour_labels: the labels of all neighbour nodes
        :param neighbourhood_activities: the node states of all neighbourhood nodes
        :param connection_states: the states of all the incoming connections of the node being processed
        :param current_activity: the state of the node after the previous timestep
        :param past_activities: the past activities of previous timesteps for all nodes
        :param input: the input for this timestep, if provided, or None otherwise
        """
        self.node_label = node_label
        self.timestep = timestep
        self.activities = activities
        self.neighbour_labels = neighbour_labels
        self.neighbourhood_activities = neighbourhood_activities
        self.connection_states = connection_states
        self.current_activity = current_activity
        self.past_activities = past_activities
        self.input = input
        self.added_nodes = []
        self.removed_nodes = []

    def activity_of(self, node_label):
        return self.activities[node_label]

    # TODO support incoming_links also; must provide one or the other
    def add_node(self, state, outgoing_links, nodel_label):
        self.added_nodes.append((state, outgoing_links, nodel_label))

    def remove_node(self, nodel_label):
        self.removed_nodes.append(nodel_label)

    def past_activity_of(self, node_label, past_activity_index=-1):
        """
        Returns the activity of the node with the given node label, in the past.
        :param node_label: the label of the node whose activity is being requested
        :param past_activity_index: the index of the past activity (-1 means the timestep before the last timestep)
        :return: the activity of the node with the given label, in the past
        """
        return self.past_activities[past_activity_index][node_label]

    def edge_data(self, neigbour_label, attr, index=0):
        return self.connection_states[neigbour_label][index][attr]


class TopologyContext(object):
    """
    The TopologyContext contains the network and its activities for a particular timestep.
    """

    __slots__ = ("network", "activities", "timestep")

    def __init__(self, network, activities, t):
        self.network = network
        self.activities = activities
        self.timestep = t


class PerturbationContext(object):
    """
    The PerturbationContext contains the node label, activity and input for a particular timestep.
    """

    __slots__ = ("node_label", "node_activity", "timestep", "input")

    def __init__(self, node_label, node_activity, timestep, input):
        self.node_label = node_label
        self.node_activity = node_activity
        self.timestep = timestep
        self.input = input


class UpdateOrder(Enum):
    ACTIVITIES_FIRST = 1
    TOPOLOGY_FIRST = 2
    SYNCHRONOUS = 3


def evolve(network, initial_conditions=None, activity_rule=None, timesteps=None, input=None, topology_rule=None,
           perturbation=None, past_conditions=None, update_order=UpdateOrder.ACTIVITIES_FIRST, copy_network=True,
           compression=False, persist_activities=True, persist_network=True):
    """
    Evolves the given network with the given initial conditions, for the specified number of timesteps, using the given
    activity and topology rules. Note that if A(t) is the network at timestep t, and S(t) is the network activity vector
    at timestep t, and F is a function defining a topology rule, and G is a function describing an activity rule, then
    A(t+1) and S(t+1) are defined as follows:
    When `update_order` is `UpdateOrder.ACTIVITIES_FIRST`:
    S(t + 1) = G(A(t), S(t))
    A(t + 1) = F(A(t), S(t + 1))
    When `update_order` is `UpdateOrder.TOPOLOGY_FIRST`:
    A(t + 1) = F(A(t), S(t))
    S(t + 1) = G(A(t + 1), S(t))
    When`update_order` is `UpdateOrder.SYNCHRONOUS`:
    A(t + 1) = F(A(t), S(t))
    S(t + 1) = G(A(t), S(t))
    :param initial_conditions: the initial activities of the network
    :param network: the network defining the topology of the system
    :param activity_rule: the rule that will determine the activity of a node in the network
    :param timesteps: the number of steps in the evolution of the network; Note that the initial state, specified by the
                      initial_conditions, is considered the result of a timestep, so that the activity_rule is invoked
                      t - 1 times; for example, if timesteps is 6, then the initial state is considered the result of
                      the first timestep (t=0), and the activity_rule will be invoked five times; the activity_rule
                      (and any perturbation) will receive the current timestep number each time it is invoked;
                      specifying the timesteps implies an automaton that evolves on its own, in the absence of any
                      external driving signal
    :param input: a list representing the inputs to the network at each timestep, or a function that accepts the
                  current timestep number, returns the input for that timestep, or None to signal the end of the
                  evolution; if a list is provided, each item in the list contains the input for each node in the
                  network for a particular timestep; the activity_rule (and any perturbation) will receive the current
                  input value for a given node through the NodeContext, when the input parameter is provided;
                  either the input or timesteps parameter must be provided, but not both; if the input parameter is
                  provided, it will override the timesteps parameter, and the timesteps parameter will have no effect;
                  the first item in the input list is given to the network at t=1, the second at t=2, etc. (i.e. no
                  input is specified for the initial state); specifying the input implies an automaton whose evolution
                  is driven by an external signal
    :param topology_rule: the rule that will determine the topology of the network
    :param perturbation: a function that defines a perturbation applied as the system evolves; the function accepts
                         one parameter: the PerturbationContext. It contains the node index, the timestep, the computed
                         activity for the node, and its input (or None if there is no input). The function must return
                         the new activity for the node at the timestep.
    :param past_conditions: a list of lists that represent activities of the network that existed before the initial
                           timestep; if this parameter is provided, then the NodeContext will contain past_activities;
                           there will be as many past_activities entries as there are entries in this list
    :param update_order: the order in which to perform the update to the system (default is
                         `UpdateOrder.ACTIVITIES_FIRST`)
    :param copy_network: whether to copy the network during its evolution (default is True)
    :param compression: whether to compress the network when it is persisted in the State (default is False)
    :param persist_activities: whether to persist the activities in the State (default is True)
    :param persist_network: whether to persist the network in the State (default is True)
    :return: a trajectory, which is a list of States, where each State defines the network and its activities for a
             particular timestep
    """
    if initial_conditions is None:
        initial_conditions = {}

    # convert initial_conditions to map, if it isn't already
    if not isinstance(initial_conditions, dict) and isinstance(initial_conditions, (list, np.ndarray)):
        initial_conditions = {i: check_np(v) for i, v in enumerate(initial_conditions)}

    input_fn = _get_input_function(timesteps, input)

    if len(initial_conditions) != len(network) and activity_rule:
        raise Exception("too few intial conditions specified [%s] for the number of given nodes [%s]" %
                        (len(initial_conditions), len(network)))

    # key is the timestep
    trajectory = {0: State(activities=initial_conditions, network=network, compression=compression)}

    prev_activities = initial_conditions
    prev_network = network

    t = 1
    while True:
        inp = input_fn(t, prev_activities, prev_network)
        if inp is None:
            break

        past = _get_past_activities(past_conditions, trajectory, t)
        curr_activities = {}
        trajectory[t] = State(compression=compression)

        if update_order is UpdateOrder.ACTIVITIES_FIRST:
            added_nodes, removed_nodes = evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
                                                           trajectory, prev_network, past,
                                                           perturbation, persist_activities)
            curr_network = evolve_topology(topology_rule, t, curr_activities, prev_network,
                                           trajectory, copy_network, persist_network, added_nodes, removed_nodes)

        elif update_order is UpdateOrder.TOPOLOGY_FIRST:
            curr_network = evolve_topology(topology_rule, t, prev_activities, prev_network,
                                           trajectory, copy_network, persist_network)
            # added and removed nodes are ignore in this case
            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
                              trajectory, curr_network, past, perturbation, persist_activities)

        elif update_order is UpdateOrder.SYNCHRONOUS:
            # TODO create test
            curr_network = evolve_topology(topology_rule, t, prev_activities, prev_network,
                                           trajectory, copy_network, persist_network)
            # added and removed nodes are ignore in this case
            evolve_activities(activity_rule, t, inp, curr_activities, prev_activities,
                              trajectory, prev_network, past, perturbation, persist_activities)

        else:
            raise Exception("unsupported update_order: %s" % update_order)

        prev_activities = curr_activities
        prev_network = curr_network

        t += 1

    # since we require Python 3.6, and dicts respect insertion order, we don't sort the trajectory entries by key
    # (even though the 3.6 language spec doesn't officially support it)
    return list(trajectory.values())


def evolve_activities(activity_rule, t, inp, curr_activities, prev_activities, trajectory, network,
                      past, perturbation, persist_activities):
    added_nodes = []
    removed_nodes = []
    if activity_rule:
        added_nodes, removed_nodes = do_activity_rule(t, inp, curr_activities, prev_activities, network,
                                                      activity_rule, past, perturbation)

        if added_nodes:
            for state, outgoing_links, node_label in added_nodes:
                curr_activities[node_label] = check_np(state)

    if persist_activities:
        trajectory[t].activities = curr_activities

    return added_nodes, removed_nodes


def do_activity_rule(t, inp, curr_activities, prev_activities, network, activity_rule, past, perturbation):
    added_nodes = []
    removed_nodes = []

    for node_label in network.nodes:
        incoming_connections = network.in_edges(node_label)
        neighbour_labels = [i for i in incoming_connections]
        current_activity = prev_activities[node_label]
        neighbourhood_activities = [prev_activities[neighbour_label] for neighbour_label in neighbour_labels]
        node_in = None if inp == "__timestep__" else inp[node_label] if _is_indexable(inp) else inp
        ctx = NodeContext(node_label, t, prev_activities, neighbour_labels, neighbourhood_activities,
                          incoming_connections, current_activity, past, node_in)

        new_activity = activity_rule(ctx)
        new_activity = check_np(new_activity)

        if ctx.added_nodes:
            added_nodes.extend(ctx.added_nodes)
        if ctx.removed_nodes:
            removed_nodes.extend(ctx.removed_nodes)

        if node_label not in ctx.removed_nodes:
            curr_activities[node_label] = new_activity

        if perturbation is not None:
            pctx = PerturbationContext(node_label, curr_activities[node_label], t, node_in)
            curr_activities[node_label] = perturbation(pctx)

    return added_nodes, removed_nodes


def evolve_topology(topology_rule, t, activities, prev_network, trajectory, copy_network,
                    persist_network, added_nodes=None, removed_nodes=None):
    network = prev_network
    if added_nodes or removed_nodes:
        network = network.copy()

        # adjust connectivity map according to node deletions and insertions
        for node_label in removed_nodes:
            network.remove_node(node_label)

        for activity, outgoing_links, node_label in added_nodes:
            network.add_node(node_label, activity=check_np(activity))
            for target, connection_state in outgoing_links.items():
                network.add_edge(node_label, target, **connection_state)

    if topology_rule:
        if copy_network:
            network = network.copy()
        network = topology_rule(TopologyContext(network, activities, t))
        if not network:
            raise Exception("topology rule must return a network")

    if persist_network:
        trajectory[t].network = network

    return network


def _get_past_activities(past_conditions, trajectory, t):
    if past_conditions and len(past_conditions) > 0:
        past_activities = [None]*len(past_conditions)
        last_t = t - 1
        for i in range(len(past_conditions)-1, -1, -1):

            curr_past_cond = past_conditions[last_t-1] if last_t < 1 else trajectory[last_t-1].activities
            if not isinstance(curr_past_cond, dict) and isinstance(curr_past_cond, (list, np.ndarray)):
                curr_past_cond = {i: v for i, v in enumerate(curr_past_cond)}

            past_activities[i] = curr_past_cond
            last_t -= 1
        return past_activities
    return None


def _get_input_function(timesteps=None, input=None):
    if timesteps is None and input is None:
        raise Exception("either the timesteps or input must be provided")

    if input is not None:
        if callable(input):
            return input
        else:
            return _ListInputFunction(input)

    return _TimestepInputFunction(timesteps)


def _is_indexable(obj):
    return isinstance(obj, collections.Sequence)


class _TimestepInputFunction:
    def __init__(self, num_steps):
        self._num_steps = num_steps

    def __call__(self, t, activities, network):
        if t == self._num_steps:
            return None
        return "__timestep__"


class _ListInputFunction:
    def __init__(self, input_list):
        self._input_list = input_list

    def __call__(self, t, activities, network):
        if (t-1) >= len(self._input_list):
            return None
        return self._input_list[t-1]


def init_simple(size, val=1, dtype=np.int):
    """
    Returns an array initialized with zeroes, with its center value set to the specified value, or 1 by default.
    :param size: the size of the array to be created
    :param val: the value to be used in the center of the array (1, by default)
    :param dtype: the data type
    :return: a vector with shape (1, size), with its center value initialized to the specified value, or 1 by default
    """
    x = np.zeros(size, dtype=dtype)
    x[len(x)//2] = val
    return x.tolist()


def init_random(size, k=2, n_randomized=None, empty_value=0, dtype=np.int):
    """
    Returns a randomly initialized array with values consisting of numbers in {0,...,k - 1}, where k = 2 by default.
    If dtype is not an integer type, then values will be uniformly distributed over the half-open interval [0, k - 1).
    :param size: the size of the array to be created
    :param k: the number of states in the Network Automaton (2, by default)
    :param n_randomized: the number of randomized sites in the array; this value must be >= 0 and <= size, if specified;
                         if this value is not specified, all sites in the array will be randomized; the randomized sites
                         will be centered in the array, while all others will have an empty value
    :param empty_value: the value to use for non-randomized sites (0, by default)
    :param dtype: the data type
    :return: a vector with shape (1, size), randomly initialized with numbers in {0,...,k - 1}
    """
    if n_randomized is None:
        n_randomized = size
    if n_randomized > size or n_randomized < 0:
        raise Exception("the number of randomized sites, if specified, must be >= 0 and <= size")
    pad_left = (size - n_randomized) // 2
    pad_right = (size - n_randomized) - pad_left
    if np.issubdtype(dtype, np.integer):
        rand_nums = np.random.randint(k, size=n_randomized, dtype=dtype)
    else:
        rand_nums = np.random.uniform(0, k - 1, size=n_randomized).astype(dtype)
    return np.array(np.pad(np.array(rand_nums), (pad_left, pad_right), 'constant', constant_values=empty_value)).tolist()


def init_simple2d(rows, cols, val=1, dtype=np.int):
    """
    Returns a list initialized with zeroes, with its center value set to the specified value, or 1 by default, when the
    list is reshaped according to the given number of rows and columns.
    :param rows: the number of rows
    :param cols: the number of columns
    :param val: the value to be used in the center of the matrix (1, by default)
    :param dtype: the data type
    :return: a list with size rows * cols, with the center value initialized to the specified value, or 1 by default
    """
    x = np.zeros((rows, cols), dtype=dtype)
    x[x.shape[0]//2][x.shape[1]//2] = val
    return np.array(x).reshape(rows * cols).tolist()


def check_np(obj):
    if isinstance(obj, np.generic):
        return np.asscalar(obj)
    return obj



from .evolution import *
from .asynchronous_rule import *
from .hopfield_net import *
from .reversible_rule import *
from .analysis import *
from .rule_tables import *
from .hexplot import *
from .sandpile import *
from .turing_machine import *
from .utils import *
from .hopfield_tank_tsp_net import *
from .substitution_system import *
from .wolfram_physics_model import *
from .fungal_growth_model import *
from .state import *
from .turtle import *
from .ctrbl_rule import *
from .langtons_loop import *

from . import topology
from . import rules
from . import vis


import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation


class Turtle:
    def __init__(self, start_x=0, start_y=0, start_orientation=0):
        """
        A simple Turtle graphics implementation using Matplotlib for graphics.

        :param start_x: the starting x-coordinate (optional, default is 0)

        :param start_y: the starting y-coordinate (optional, default is 0)

        :param start_orientation: the starting orientation in degrees (optional, default is 0)
        """
        self.x = start_x
        self.y = start_y
        self.orientation = start_orientation
        self.trajectory = []
        self._stack = []

    def forward(self, distance=1):
        """
        Move the Turtle forward, from the current orientation, by the given amount.

        :param distance: the amount by which to move the Turtle (default is 1)
        """
        angle = math.radians(self.orientation)
        new_x = math.sin(angle) * distance + self.x
        new_y = math.cos(angle) * distance + self.y
        p1 = [self.x, self.y]
        p2 = [new_x, new_y]
        x_values = [p1[0], p2[0]]
        y_values = [p1[1], p2[1]]
        self.trajectory.append((x_values, y_values))
        self.x = new_x
        self.y = new_y

    def rotate(self, degrees):
        """
        Rotate the Turtle by the given number of degrees. A positive number turns right, and a negative number
        turns left.

        :param degrees: the amount by which to turn, in degrees
        """
        self.orientation += degrees

    def push(self):
        """
        Pushes the current state (x- and y-coordinates, and orientation) onto an internal stack.
        """
        self._stack.append((self.x, self.y, self.orientation))

    def pop(self):
        """
        Pops the last stored state (x- and y-coordinates, and orientation) from the internal stack, and sets
        the state of the Turtle to that state.
        """
        x, y, orientation = self._stack.pop()
        self.x = x
        self.y = y
        self.orientation = orientation

    @property
    def pos(self):
        """
        Return the current position of the Turtle.

        :return: a 2-tuple, representing the x and y coordinates of the Turtle
        """
        return self.x, self.y

    @property
    def abs(self):
        """
        The magnitude of the vector describing the current position of the Turtle (i.e. the distance from the origin).

        :return: the absolute value of the position vector
        """
        return (self.x ** 2 + self.y ** 2) ** 0.5

    def plot(self, marker=""):
        fig, ax = plt.subplots()
        ax.set_aspect('equal')
        for state in self.trajectory:
            plt.plot(state[0], state[1], marker=marker, color="black")

        plt.gca().axes.xaxis.set_ticklabels([])
        plt.gca().axes.yaxis.set_ticklabels([])
        plt.xticks([])
        plt.yticks([])
        plt.show()

    def animate(self, save=False, interval=50, dpi=80, marker="", repeat=False, fps=50):
        fig, ax = plt.subplots()
        ax.set_aspect('equal')

        x_vals = [s[0] for s in self.trajectory]
        max_x = np.max(x_vals)
        min_x = np.min(x_vals)
        max_x += 0.05*(max_x - min_x)
        min_x -= 0.05*(max_x - min_x)
        y_vals = [s[1] for s in self.trajectory]
        max_y = np.max(y_vals)
        min_y = np.min(y_vals)
        max_y += 0.05*(max_y - min_y)
        min_y -= 0.05*(max_y - min_y)

        def update(t):
            plt.plot(self.trajectory[t][0], self.trajectory[t][1], marker=marker, color="black")
            if t == len(self.trajectory) - 1 and repeat:
                ax.clear()
                ax.xaxis.set_ticklabels([])
                ax.yaxis.set_ticklabels([])
                ax.set_xticks([])
                ax.set_yticks([])
                plt.xlim(min_x, max_x)
                plt.ylim(min_y, max_y)

        ani = animation.FuncAnimation(fig, update, frames=len(self.trajectory), interval=interval,
                                      save_count=len(self.trajectory), repeat=repeat)

        plt.xlim(min_x, max_x)
        plt.ylim(min_y, max_y)

        plt.gca().axes.xaxis.set_ticklabels([])
        plt.gca().axes.yaxis.set_ticklabels([])
        plt.xticks([])
        plt.yticks([])

        if save:
            ani.save('turtle.gif', dpi=dpi, writer="imagemagick", fps=fps)
        plt.show()



import networkx as nx
import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib.collections as mcoll
import collections
from .state import State
from .substitution_system import SubstitutionSystem
try:
    import cPickle as pickle
except:
    import pickle


def plot_activities(trajectory_or_activities, shape=None, slice=-1, title='', colormap='Greys', vmin=None, vmax=None,
                    node_annotations=None, show_grid=False):
    if len(trajectory_or_activities) is 0:
        raise Exception("there are no activities")
    if isinstance(trajectory_or_activities[0], State):
        # trajectory_or_activities is a trajectory, convert it to an activities list
        activities = get_activities_over_time_as_list(trajectory_or_activities)
    else:
        activities = trajectory_or_activities
    if shape is not None:
        activities = _reshape_for_animation(activities, shape)
    plot_grid(activities, shape, slice, title, colormap, vmin, vmax, node_annotations, show_grid)


def get_activities_over_time_as_list(trajectory):
    activities = []
    for state in trajectory:
        ac = state.activities
        row = []
        if ac:
            for a in sorted(ac):
                row.append(ac[a])
            activities.append(row)
    return activities


def plot_grid(activities, shape=None, slice=-1, title='', colormap='Greys', vmin=None, vmax=None,
              node_annotations=None, show_grid=False):
    if shape is not None:
        activities = np.array(activities).reshape((len(activities), shape[0], shape[1]))[slice]
    cmap = plt.get_cmap(colormap)
    plt.title(title)
    plt.imshow(activities, interpolation='none', cmap=cmap, vmin=vmin, vmax=vmax)

    if node_annotations is not None:
        for i in range(len(node_annotations)):
            for j in range(len(node_annotations[i])):
                plt.text(j, i, node_annotations[i][j], ha="center", va="center", color="grey",
                         fontdict={'weight':'bold','size':6})

    if show_grid:
        plt.grid(which='major', axis='both', linestyle='-', color='grey', linewidth=0.5)
        plt.xticks(np.arange(-.5, len(activities[0]), 1), "")
        plt.yticks(np.arange(-.5, len(activities), 1), "")
        plt.tick_params(axis='both', which='both', length=0)

    plt.show()


def plot_grid_multiple(ca_list, shape=None, slice=-1, titles=None, colormap='Greys', vmin=None, vmax=None):
    cmap = plt.get_cmap(colormap)
    for i in range(0, len(ca_list)):
        plt.figure(i)
        if titles is not None:
            plt.title(titles[i])
        activities = list(ca_list[i])
        if shape is not None:
            activities = np.array(activities).reshape((len(activities), shape[0], shape[1]))[slice]
        plt.imshow(activities, interpolation='none', cmap=cmap, vmin=vmin, vmax=vmax)
    plt.show()


def animate_activities(trajectory_or_activities, title='', shape=None, save=False, interval=50, colormap='Greys',
                       vmin=None, vmax=None, show_grid=False, show_margin=True, scale=0.6, dpi=80):
    if len(trajectory_or_activities) is 0:
        raise Exception("there are no activities")
    if isinstance(trajectory_or_activities[0], State):
        # trajectory_or_activities is a trajectory, convert it to an activities list
        activities = get_activities_over_time_as_list(trajectory_or_activities)
    else:
        activities = trajectory_or_activities
    if shape is not None:
        activities = _reshape_for_animation(activities, shape)
    cmap = plt.get_cmap(colormap)
    fig, ax = plt.subplots()
    plt.title(title)
    if not show_margin:
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)

    grid_linewidth = 0.0
    if show_grid:
        plt.xticks(np.arange(-.5, len(activities[0][0]), 1), "")
        plt.yticks(np.arange(-.5, len(activities[0]), 1), "")
        plt.tick_params(axis='both', which='both', length=0)
        grid_linewidth = 0.5

    vertical = np.arange(-.5, len(activities[0][0]), 1)
    horizontal = np.arange(-.5, len(activities[0]), 1)
    lines = ([[(x, y) for y in (-.5, horizontal[-1])] for x in vertical] +
             [[(x, y) for x in (-.5, vertical[-1])] for y in horizontal])
    grid = mcoll.LineCollection(lines, linestyles='-', linewidths=grid_linewidth, color='grey')
    ax.add_collection(grid)

    im = plt.imshow(activities[0], animated=True, cmap=cmap, vmin=vmin, vmax=vmax)
    if not show_margin:
        baseheight, basewidth = im.get_size()
        fig.set_size_inches(basewidth*scale, baseheight*scale, forward=True)

    i = {'index': 0}
    def updatefig(*args):
        i['index'] += 1
        if i['index'] == len(activities):
            i['index'] = 0
        im.set_array(activities[i['index']])
        return im, grid
    ani = animation.FuncAnimation(fig, updatefig, interval=interval, blit=True, save_count=len(activities))
    if save:
        ani.save('evolved.gif', dpi=dpi, writer="imagemagick")
    plt.show()


def animate_plot1D(x, y, save=False, interval=50, dpi=80):
    if isinstance(y[0], State):
        # y is a trajectory, convert it to an activities list
        y = get_activities_over_time_as_list(y)
    fig1 = plt.figure()
    line, = plt.plot(x, y[0])
    def update_line(activity):
        line.set_data(x, activity)
        return line,
    ani = animation.FuncAnimation(fig1, update_line, frames=y, blit=True, interval=interval)
    if save:
        ani.save('plot.gif', dpi=dpi, writer="imagemagick")
    plt.show()


def _reshape_for_animation(activities, shape):
    if len(shape) == 1:
        assert shape[0] == len(activities[0]), "shape must equal the length of an activity vector"
        new_activities = []
        for i, a in enumerate(activities):
            new_activity = []
            new_activity.extend(activities[0:i+1])
            while len(new_activity) < len(activities):
                new_activity.append([0]*len(activities[0]))
            new_activities.append(new_activity)
        return np.array(new_activities)
    elif len(shape) == 2:
        return np.reshape(activities, (len(activities), shape[0], shape[1]))
    else:
        raise Exception("shape must be a tuple of length 1 or 2")


def plot_network(network, layout="shell", with_labels=True, node_color="#1f78b4", node_size=300):
    G = network.to_networkx()
    if layout == "shell":
        nx.draw_shell(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
    elif layout == "spring":
        nx.draw_spring(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
    elif isinstance(layout, dict):
        nx.draw(G, pos=layout, with_labels=with_labels, node_color=node_color, node_size=node_size)
    else:
        raise Exception("unsupported layout: %s" % layout)
    plt.show()


def animate_network(trajectory, save=False, interval=50, dpi=80, layout="shell",
                    with_labels=True, node_color="b", node_size=30):
    fig, ax = plt.subplots()

    def update(state):
        ax.clear()

        network = state.network
        G = nx.MultiDiGraph()
        for n in network.nodes:
            G.add_node(n)
        for edge in network.edges:
            G.add_edge(edge[0], edge[1])

        if layout == "shell":
            nx.draw_shell(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
        elif layout == "spring":
            nx.draw_spring(G, with_labels=with_labels, node_color=node_color, node_size=node_size)
        elif isinstance(layout, dict):
            nx.draw(G, pos=layout, with_labels=with_labels, node_color=node_color, node_size=node_size)
        else:
            raise Exception("unsupported layout: %s" % layout)

    ani = animation.FuncAnimation(fig, update, frames=trajectory, interval=interval,
                                  save_count=len(trajectory))
    if save:
        ani.save('evolved.gif', dpi=dpi, writer="imagemagick")
    plt.show()


def plot1D(x, y, color=None, label=None, xlabel=None, ylabel=None, xlim=None, ylim=None, twinx=False,
           legend=None, tight_layout=None, title=None):
    """
    Creates a 1D plot of the given x and y values.

    :param x: A list representing the values of the x-axis.

    :param y: A list representing the values of the y-axis. If the values of this list are also lists, then each will
              be plotted as a separate series on the y-axis.

    :param color: A string, representing the color of the series, or a list of colors, representing the color of each
                   series to be plotted. The number of colors must match the number of series. The color values must
                   be recognizable by Matplotlib.

    :param label: A string or number, or a list of strings or numbers, representing the labels of each series to be
                  plotted. The number of labels must match the number of series.

    :param xlabel: A string representing the label of the x-axis.

    :param ylabel: A string, or list of at most two strings, representing the label(s) of the y-axis.

    :param xlim: A 2-tuple of numbers, representing the limits of the x-axis.

    :param ylim: A 2-tuple of numbers, or a list of at most two 2-tuples of numbers, representing the
                 limits of the y-axis.

    :param twinx: If True, the provided y series will be plotted on two separate y-axes, and there must
                  be at least two series of y values provided. If there are more than two series of y values,
                  then each series will be plotted on alternating y-axes. (Default is False)

    :param legend: A dict, or a list of at most two dicts, representing the arguments to the legend
                   function of Matplotlib.

    :param tight_layout: A dict representing the arguments to the tight_layout function of Matplotlib.

    :param title: The plot title.
    """
    axes = []
    fig, ax1 = plt.subplots()
    axes.append(ax1)
    if twinx:
        axes.append(ax1.twinx())
        if len(y) < 2:
            raise Exception("there must be at least two series of y values provided")
        for _y in y:
            if not isinstance(_y, (list, np.ndarray)):
                raise Exception("an item in y must be a list representing a y series")

    current_axis_idx = 0
    # assume that the first value type in y is representative of all the value types in y
    is_multiseries = isinstance(y[0], (list, np.ndarray))

    def _get_item(items, i):
        if isinstance(items, str):
            return items
        if isinstance(items, collections.Sequence):
            return items[i]
        return items

    if not is_multiseries:
        y = [y]

    for i, y_series in enumerate(y):
        current_axis = axes[current_axis_idx % len(axes)]
        plot_args = {}
        if label:
            plot_args["label"] = _get_item(label, i)
        if color:
            plot_args["color"] = _get_item(color, i)
        current_axis.plot(x, y_series, **plot_args)
        if xlabel:
            current_axis.set_xlabel(xlabel)
        if ylabel:
            ylabel_args = {}
            if color:
                ylabel_args["color"] = _get_item(color, i)
            current_axis.set_ylabel(_get_item(ylabel, i % 2), **ylabel_args)
        if xlim:
            current_axis.set_xlim(xlim)
        if ylim:
            current_axis.set_ylim(_get_item(ylim, i % 2))
        if legend:
            current_axis.legend(**legend)

        current_axis_idx += 1

    if tight_layout:
        plt.tight_layout(**tight_layout)

    if title:
        plt.title(title)

    plt.show()


def poincare_plot(activities, timesteps, xlabel=None, ylabel=None, xlim=None, ylim=None, title=None):
    """
    Create a Poincaré plot.

    :param activities: A list of activities. If the values of this list are also lists, then each will
                       be plotted as a separate series.

    :param timesteps: The number of timesteps in the trajectory to consider, starting from the end.

    :param xlabel: A string representing the label of the x-axis.

    :param ylabel: A string representing the label of the y-axis.

    :param xlim: A 2-tuple of numbers, representing the limits of the x-axis.

    :param ylim: A 2-tuple of numbers, or a list of at most two 2-tuples of numbers, representing the
                 limits of the y-axis.

    :param title: The plot title.
    """
    cm = plt.get_cmap('gist_rainbow')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    # assume that the first value type in y is representative of all the value types in y
    is_multiseries = isinstance(activities[0], (list, np.ndarray))
    if is_multiseries:
        ax.set_prop_cycle(color=[cm(1. * i / len(activities)) for i in range(len(activities))])
    else:
        activities = [activities]
    for a in activities:
        x = []
        y = []
        for t in range(timesteps - 1):
            x.append(a[-timesteps:][t])
            y.append(a[-timesteps:][t + 1])
        plt.scatter(x, y, s=1)
    if xlim:
        plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)
    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)
    plt.show()


def bifurcation_plot(x, trajectories, timesteps, xlabel=None, ylabel=None, title=None):
    """
    Create a bifurcation plot.

    :param x: A list representing the values of the x-axis.

    :param trajectories: A list of lists. Each inner list is a sequence of activities representing a trajectory.

    :param timesteps: The number of timesteps in the trajectory to consider, starting from the end.

    :param xlabel: A string representing the label of the x-axis.

    :param ylabel: A string representing the label of the y-axis.

    :param title: The plot title.
    """
    y = []
    for t in trajectories:
        y.append(np.unique(t[-timesteps:]))
    for x_e, y_e in zip(x, y):
        plt.scatter([x_e] * len(y_e), y_e, color="b", s=1)
    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)
    plt.show()


def plot_degree_distribution(network, xlabel="Node degree", ylabel_freq="Frequency", ylabel_prob="Probability",
                             in_degree=False, out_degree=False,
                             equation=None, equation_x=0.51, equation_y=0.76, equation_text="", equation_color="r",
                             color="r", title=None):
    """
    Create a node degree distribution plot for the given network.

    :param network: A Netomaton Network instance.

    :param xlabel: The x-axis label.

    :param ylabel_freq: The frequency y-axis label.

    :param ylabel_prob: The probability y-axis label.

    :param in_degree: If True, the in-degree will be used. (default is False)

    :param out_degree: If True, the out-degree will be used. (default is False)

    :param equation: A callable that computes the degree distribution, given a node degree.

    :param equation_x: The equation's x coordinate.

    :param equation_y: The equation's y coordinate.

    :param equation_text: The equation to display.

    :param equation_color: The equation text's color. It must be a valid Matplotlib color.

    :param color: The color to use for the plot. It must be a valid Matplotlib color.

    :param title: The plot's title.
    """
    degree_counts = {}
    for node in network.nodes:
        if in_degree:
            degree = network.in_degree(node)
        elif out_degree:
            degree = network.out_degree(node)
        else:
            degree = network.degree(node)

        if degree not in degree_counts:
            degree_counts[degree] = 0
        degree_counts[degree] += 1

    x = [i for i in range(1, max(degree_counts) + 1)]
    height = [degree_counts[i] if i in degree_counts else 0 for i in x]
    plt.bar(x, height)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel_freq)

    if equation:
        y = [equation(k) for k in x]
        plt.twinx()
        plt.plot(x, y, color=color)
        plt.ylabel(ylabel_prob)
        plt.text(equation_x, equation_y, equation_text, transform=plt.gca().transAxes, color=equation_color)

    if title:
        plt.title(title)

    plt.show()


def ncr(n, r):
    """
    Returns the number of ways to choose r items from n items without repetition and without order.

    :param n: A non-negative integer.

    :param r: A non-negative integer.

    :return: An integer representing the number of ways to choose r items from n items without repetition
             and without order.
    """
    if not isinstance(n, int) or n < 0:
        raise Exception("n must be a non-negative int")
    if not isinstance(r, int) or r < 0:
        raise Exception("r must be a non-negative int")
    if n < r:
        return 0
    f = math.factorial
    return f(n) // f(r) // f(n-r)


def plot_L_system(state, turtle, bindings, **kwargs):
    _render_L_system(state, bindings)
    turtle.plot(**kwargs)


def animate_L_system(state, turtle, bindings, **kwargs):
    _render_L_system(state, bindings)
    turtle.animate(**kwargs)


def _render_L_system(state, bindings):
    state = SubstitutionSystem.to_string([state])[0]
    for s in state:
        instructions = bindings[s]
        if not isinstance(instructions, list):
            instructions = [instructions]
        for instruction in instructions:
            if callable(instruction):
                instruction()
            else:
                fn, *args = instruction
                fn(*args)


def binarize_for_plotting(activities):
    """
    Expects a list of lists of single numbers representing the activities (e.g. [[2], [35], [12]]),
    and returns a list of lists of the binarized versions of the given numbers, left-padded with
    zeroes. (e.g. [[0, 0, 0, 0, 1, 0], [1, 0, 0, 0, 1, 1], [0, 0, 1, 1, 0, 0]])

    :param activities: the activities as a list of lists of single numbers

    :return: a list of lists of the binary versions of the numbers, left-padded with zeroes
    """
    activities = [[int(x) for x in bin(int(a[0]))[2:]] for a in activities]
    max_len = np.max([len(a) for a in activities])
    return np.asarray([np.pad(a, (max_len - len(a), 0), 'constant', constant_values=0) for a in activities])



from .topology import cellular_automaton2d


class CTRBLRule:
    """
    A particular kind of rule that operates on von Neumann neighbourhoods, taking into account the states of a cell's
    neighbours at the top, bottom, left and right positions explicitly. Only 2D automata with periodic boundaries and a
    radius of 1 are supported.
    """
    def __init__(self, dim, rule_table, add_rotations=False):
        """
        Creates a CTRBLRule instance.

        :param dim: a 2-tuple, representing the rows and columns in the 2D Cellular Automata

        :param rule_table: a dictionary with keys being a 5-tuple representing the states of the CTRBL cells, and values
                           being a single value representing the image state (i.e. the state of the Center cell in
                           the next timestep); all combinations of states must exist, otherwise, if the combination of
                           states does not exist in the rule table, an exception will be raised

        :param add_rotations: whether rotations in the rule table are implied, and should be included (default is False)
        """
        self._network = cellular_automaton2d(dim[0], dim[1], r=1, neighbourhood="von Neumann")
        self._neighbourhood_map = self._init_neighbourhood_map(dim)
        self._rule_table = self._init_rule_table(rule_table, add_rotations)

    def activity_rule(self, ctx):
        neighbourhood = self._neighbourhood_map[ctx.node_label]
        c = ctx.current_activity
        t = ctx.activity_of(neighbourhood.top)
        r = ctx.activity_of(neighbourhood.right)
        b = ctx.activity_of(neighbourhood.bottom)
        l = ctx.activity_of(neighbourhood.left)
        key = (c, t, r, b, l)
        if key not in self._rule_table:
            raise Exception("neighbourhood state (%s, %s, %s, %s, %s) not in rule table" % key)
        return self._rule_table[key]

    @property
    def network(self):
        return self._network

    @property
    def neighbourhood_map(self):
        return self._neighbourhood_map

    @property
    def rule_table(self):
        return self._rule_table

    @staticmethod
    def _init_neighbourhood_map(dim):
        """
        Returns a map of the cell label to a dictionary of the labels of the top, bottom, left, and right neighbours.
        e.g. {61: VonNeumannNeighbourhood(top=1, bottom=121, left=60, right=62, center=61)}

        :param dim: a 2-tuple, representing the rows and columns in the 2D Cellular Automata

        :return: the neighbourhood map
        """
        rows, cols = dim
        neighbourhood_map = {}
        idx = 0
        for row in range(rows):
            for _ in range(cols):
                center = idx
                left = (((center-row*cols) - 1) % cols) + (row*cols)
                right = (((center-row*cols) + 1) % cols) + (row*cols)
                top = (center - cols) % (rows*cols)
                bottom = (center + cols) % (rows*cols)
                neighbourhood_map[idx] = VonNeumannNeighbourhood(center=center, top=top, bottom=bottom,
                                                                 left=left, right=right)
                idx += 1
        return neighbourhood_map

    @staticmethod
    def _init_rule_table(rule_table, add_rotations):
        new_rule_table = {}
        for rule, image in rule_table.items():
            new_rule_table[rule] = image
            if add_rotations:
                r = list(rule)
                for _ in range(3):
                    r.insert(1, r.pop(4))
                    new_rule_table[tuple(r)] = image
        return new_rule_table


class VonNeumannNeighbourhood:
    """
    A container for the labels of a von Neumann neighbourhood.
    """
    def __init__(self, center, top, bottom, left, right):
        self._center = center
        self._top = top
        self._bottom = bottom
        self._left = left
        self._right = right

    @property
    def center(self):
        return self._center

    @property
    def bottom(self):
        return self._bottom

    @property
    def top(self):
        return self._top

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    def __eq__(self, other):
        return self.center == other.center and self.top == other.top and self.bottom == other.bottom and \
               self.left == other.left and self.right == other.right

    def __repr__(self):
        return "VonNeumannNeighbourhood(center=%s, top=%s, bottom=%s, left=%s, right=%s)" % \
               (self.center, self.top, self.bottom, self.left, self.right)



import numpy as np
from .rules import *


class AsynchronousRule:
    """
    Creates an asynchronous automaton rule with a cyclic update scheme. Also known as a sequential cellular
    automaton rule, in NKS. This rule wraps a given rule, making the given rule asynchronous. To learn more about
    the details of sequential cellular automata, see the Chapter 9, Section 10 Notes on Sequential cellular
    automata in NKS.
    """
    def __init__(self, activity_rule, update_order=None, n=None, randomize_each_cycle=False):
        """
        Constructs an asynchronous rule out of a given rule. Either the `update_order` or `n` parameter must be
        specified. If no `update_order` is given, then the `n` parameter must be specified, and an update order
        list will be constructed and shuffled.
        :param activity_rule: the rule that will be made asynchronous
        :param update_order: a list containing the indices of the nodes in the Network Automaton, specifying the update
                             order; only the nodes specified in the list will be updated
        :param n: the total number of nodes in the Network Automaton
        :param randomize_each_cycle: whether to shuffle the update order list after each complete cycle
        """
        if update_order is None and n is None:
            raise Exception("either update_order or n must be specified")
        self._activity_rule = activity_rule
        if update_order is not None:
            self._update_order = update_order
        else:
            self._shuffle_update_order(n)
        self._curr = 0
        self._num_applied = 0
        self._randomize_each_cycle = randomize_each_cycle

    def _shuffle_update_order(self, n):
        self._update_order = np.arange(n)
        np.random.shuffle(self._update_order)
        self._update_order = self._update_order.tolist()

    def __call__(self, ctx):
        if ctx.node_label in self._update_order:
            self._num_applied += 1
        if not self._should_update(ctx.node_label):
            self._check_for_end_of_cycle()
            return ctx.current_activity
        self._check_for_end_of_cycle()
        return self._activity_rule(ctx)

    def _should_update(self, c,):
        return c == self._update_order[self._curr]

    def _check_for_end_of_cycle(self):
        if self._num_applied == len(self._update_order):
            self._curr = (self._curr + 1) % len(self._update_order)
            self._num_applied = 0
            if self._randomize_each_cycle:
                self._shuffle_update_order(len(self._update_order))



from .state import Network


class WolframPhysicsModel:
    def __init__(self, configuration, rules):
        self.network, self.last_node = self._init_network_and_conditions(configuration)
        self.rules = rules

    @staticmethod
    def _init_network_and_conditions(config):
        """
        The network for the Wolfram Physics Model is comprised of connection states that consist of a "label",
        that indicates the relation label, and optionally a "hyperedge_index", if the pair of nodes belong to a
        hyperedge, and a "unary" flag, to indicate a unary connection.
        """
        network = Network()
        last_node = 0
        for c, relation in enumerate(config):
            prev = None
            edge_label = str(c+1)
            is_hyperedge = len(relation) > 2
            next_edge_num = 0
            for node in relation:
                if prev is not None:
                    prev_state = {"label": edge_label}
                    if is_hyperedge:
                        prev_state["hyperedge_index"] = next_edge_num
                    network.add_edge(prev, node, **prev_state)
                    next_edge_num += 1
                prev = node
                if node > last_node:
                    last_node = node
            if len(relation) == 1:
                n = relation[0]
                network.add_edge(n, n, label=edge_label, unary=True)
        return network, last_node

    @staticmethod
    def network_to_config(network):
        edges = {}
        hyperedges = {}
        for to_node in network.nodes:
            in_edges = network.in_edges(to_node)
            for i in in_edges:
                from_node = i
                # connection = in_edges[i]
                for connection in in_edges[i]:
                    edge_label = connection["label"]
                    if "unary" in connection:
                        edges[edge_label] = (from_node,)
                    elif "hyperedge_index" in connection:
                        if edge_label not in hyperedges:
                            hyperedges[edge_label] = []
                        hyperedges[edge_label].append(((from_node, to_node), connection["hyperedge_index"]))
                    else:
                        edges[edge_label] = (from_node, to_node)

        # merge hyperedges
        merged_hyperedges = {}
        if hyperedges:
            for edge_label in hyperedges:
                parts = hyperedges[edge_label]
                merged = [0] * (len(parts) + 1)
                for part in parts:
                    edge = part[0]
                    index = part[1]
                    merged[index] = edge[0]
                    if (index + 2) == len(merged):
                        merged[-1] = edge[1]
                merged_hyperedges[edge_label] = tuple(merged)

        edges.update(merged_hyperedges)
        config = [edges[label] for label in sorted(edges)]  # sort so that we can preserve the order of the relations

        return config

    def topology_rule(self, cctx):
        relations = self.network_to_config(cctx.network)

        matched_relations = []
        unmatched_relations = []
        partially_matched = []

        # keep scanning until there is an iteration where no matches are produced
        #  reset the symbol bindings during each scan
        scanning = True
        while scanning:

            # bind nodes to symbols
            symbol_bindings = {}

            for rule in self.rules["in"]:

                match_found = False

                for relation in list(relations):  # iterate over a copy of the list so that we can remove from the original
                    if self._matches(rule, relation, symbol_bindings):
                       partially_matched.append(relation)
                       relations.remove(relation)
                       match_found = True
                       break

                if len(partially_matched) == len(self.rules["in"]):
                    matched_relations.append((partially_matched, symbol_bindings))
                    partially_matched = []

                scanning = match_found

            # move all partially matched to unmatched
            unmatched_relations.extend(partially_matched)
            partially_matched = []

        new_config = []
        # add the relations that did not match the rules
        new_config.extend(relations)
        new_config.extend(unmatched_relations)
        # add the new relations
        for matched, bindings in matched_relations:
            for rule_out in self.rules["out"]:
                new_edge = []
                for symbol in rule_out:
                    if symbol not in bindings:
                        self.last_node += 1
                        bindings[symbol] = self.last_node
                    new_edge.append(bindings[symbol])
                new_config.append(tuple(new_edge))

        self.network, self.last_node = self._init_network_and_conditions(new_config)

        return self.network

    def _matches(self, rule, relation, symbol_bindings):
        if len(rule) != len(relation):
            #  rule does not match this relation
            return False

        added_symbols = []
        for i, symbol in enumerate(rule):
            if symbol in symbol_bindings:
                if relation[i] != symbol_bindings[symbol]:
                    #  rule does not match this relation
                    for added in added_symbols:
                        del symbol_bindings[added]
                    return False
                else:
                    continue

            symbol_bindings[symbol] = relation[i]
            added_symbols.append(symbol)

        return True

    def to_configurations(self, trajectory):
        configs = []
        for state in trajectory:
            configs.append(self.network_to_config(state.network))
        return configs



import numpy as np


def cellular_automaton(n, r=1, boundary="periodic"):
    """
    Creates the adjacency matrix for a 1D cellular automaton, with the given number of cells and neighbourhood
    radius.

    :param n: the number of cells in this automaton

    :param r: the neighbourhood radius

    :param boundary: the boundary condition

    :return: an adjacency matrix describing this cellular automaton
    """
    if n < 3:
        raise Exception("There must be at least 3 cells")
    adjacency_matrix = [[0. for j in range(n)] for i in range(n)]
    if boundary == "periodic":
        for w, row in enumerate(adjacency_matrix):
            for c, _ in enumerate(row):
                if w == c:
                    adjacency_matrix[w][c] = 1.
                    for i in range(r):
                        adjacency_matrix[w][c - (1 + i)] = 1.
                        adjacency_matrix[w][(c + (1 + i)) % len(adjacency_matrix[w])] = 1.
    else:
        raise Exception("unsupported boundary condition: %s" % boundary)
    return adjacency_matrix


def cellular_automaton2d(rows, cols, r=1, neighbourhood='Moore', boundary="periodic"):
    """
    Creates the adjacency matrix for a 2D cellular automaton, with the given number of rows, columns, neighbourhood
    radius, and neighbourhood type.

    :param rows: the number of rows in the 2D automaton

    :param cols: the number of columns in the 2D automaton

    :param r: the neighbourhood radius; the neighbourhood dimensions will be 2r+1 x 2r+1

    :param neighbourhood: the neighbourhood type; valid values are 'Moore', 'von Neumann', and 'Hex'

    :param boundary: the boundary condition

    :return: an adjacency matrix describing this cellular automaton
    """
    n = rows * cols
    if n < 9:
        raise Exception("There must be at least 9 cells")
    adjacency_matrix = [[0. for j in range(n)] for i in range(n)]
    if boundary == "periodic":
        if neighbourhood == 'von Neumann':
            criteria = lambda a_i, b_i, a_o, b_o, radius, rownum: np.abs(a_i - a_o) + np.abs(b_i - b_o) <= radius
        elif neighbourhood == 'Moore':
            criteria = lambda a_i, b_i, a_o, b_o, radius, rownum: np.abs(a_i - a_o) <= radius and np.abs(b_i - b_o) <= radius
        elif neighbourhood == 'Hex':
            def hex_crit(a_i, b_i, a_o, b_o, radius, rownum):
                vn = np.abs(a_i - a_o) + np.abs(b_i - b_o) <= radius
                if rownum % 2 == 0:
                    ex = (b_i - b_o) < radius
                else:
                    ex = (b_o - b_i) < radius
                return vn or ex
            criteria = hex_crit
        else:
            raise Exception("neighbourhood type not supported: %s" % neighbourhood)

        lattice = np.array(range(n)).reshape((rows, cols)).tolist()
        rownum = 0
        for a, row in enumerate(lattice):
            rownum += 1
            for b, _ in enumerate(row):
                adjacency_row_num = lattice[a][b]
                neighbourhood_points = _get_neighbourhood_points2d(a, b, r, criteria, rownum)
                for point in neighbourhood_points:
                    x = point[0] if point[0] == -1 else point[0] % len(lattice)
                    y = point[1] if point[1] == -1 else point[1] % len(lattice[a])
                    adjacency_matrix[adjacency_row_num][lattice[x][y]] = 1.

    else:
        raise Exception("unsupported boundary condition: %s" % boundary)
    return adjacency_matrix


def _get_neighbourhood_points2d(a, b, r, criteria, rownum):
    neighbourhood_points = [[(x, y) for y in range(b-r, b+r+1)] for x in range(a-r, a+r+1)]
    filtered = [[p for p in row if criteria(p[0], p[1], a, b, r, rownum)] for row in neighbourhood_points]
    return [item for sublist in filtered for item in sublist]  # flatten the list of lists



import msgpack
import gc
import networkx as nx


class Network:
    __slots__ = "_network"

    def __init__(self, n=0):
        """
        Constructs a fully disconnected Network with n nodes.

        :param n: the number of nodes
        """
        self._network = {i: self._new_node() for i in range(n)}

    def add_edge(self, i, j, **attr):
        """
        Add an edge from i to j. If i and j do not exist, create them.
        An additional edge will be added if an edge from i to j already exists.

        :param i:

        :param j:

        :param attr:

        :return:
        """
        self._init_node(i)
        self._init_node(j)
        if i not in self._network[j]["incoming"]:
            self._network[j]["incoming"][i] = []
        self._network[j]["incoming"][i].append(attr)
        self._increment_in_degree(j)
        self._increment_out_degree(i)
        self._add_outgoing(i, j)

    def add_node(self, node_label, **attr):
        """
        Add a node. If the node already exists

        :param node_label:

        :param attr:

        :return:
        """
        self._init_node(node_label)
        self._network[node_label].update(attr)

    def update_edge(self, i, j, indices=None, **attr):
        if j not in self._network or i not in self._network[j]["incoming"]:
            raise Exception("edge does not exist")
        if indices is None:
            # update all edges in this connection
            for edge in self._network[j]["incoming"][i]:
                edge.update(attr)
        else:
            for k in indices:
                self._network[j]["incoming"][i][k].update(attr)

    def remove_node(self, node_label):
        for i, v in self._network[node_label]["incoming"].items():
            self._decrement_out_degree(i, n=len(v))
            self._remove_outgoing(i, node_label)
        for j in self._network[node_label]["outgoing"]:
            self._decrement_in_degree(j)
            self._remove_incoming(node_label, j)
        del self._network[node_label]

    def remove_edge(self, i, j):
        """
        Removes all instances of an edge from i to j.

        :param i: the originating node of the edge

        :param j: the target destination node of the edge
        """
        n = len(self._network[j]["incoming"][i])
        self._remove_outgoing(i, j)
        self._decrement_out_degree(i, n=n)
        self._decrement_in_degree(j, n=n)
        self._remove_incoming(i, j)

    def in_edges(self, node_label):
        return self._network[node_label]["incoming"]

    def edge(self, i, j):
        return self._network[j]["incoming"][i]

    def to_dict(self):
        return self._network

    @property
    def nodes(self):
        return self._network.keys()

    @property
    def edges(self):
        for j, v in self._network.items():
            for i in v["incoming"]:
                for k in v["incoming"][i]:
                    yield i, j, k

    def has_edge(self, i, j):
        return j in self._network and i in self._network[j]["incoming"]

    def degree(self, node_label):
        return self._network[node_label]["out"] + self._network[node_label]["in"]

    def in_degree(self, node_label):
        return self._network[node_label]["in"]

    def out_degree(self, node_label):
        return self._network[node_label]["out"]

    def copy(self):
        return self.decompress(self.compress())

    def compress(self):
        """
        Compresses the Network into a binary format.

        :return: packed bytes
        """
        return _Compressor.compress(self._network)

    @staticmethod
    def decompress(packed):
        """
        Decompresses a compressed Network.

        :param packed: packed bytes

        :return: a Network
        """
        n = Network()
        n._network = _Compressor.decompress(packed)
        return n

    @staticmethod
    def from_networkx(G):
        """
        Creates a Network from the given NetworkX graph.

        :param G: a NetworkX graph

        :return: a Network
        """
        network = Network()
        for node in G.nodes(data=True):
            network.add_node(node[0], **node[1])
        for edge in G.edges(data=True):
            network.add_edge(edge[0], edge[1], **edge[2])
        return network

    def to_networkx(self):
        """
        Returns this Network as a NetworkX MultiDiGraph.

        :return: a NetworkX MultiDiGraph
        """
        G = nx.MultiDiGraph()
        for node in self.nodes:
            atts = {}
            for key in self._network[node]:
                if key not in self._new_node():
                    atts[key] = self._network[node][key]
            G.add_node(node, **atts)
        for i, j, data in self.edges:
            G.add_edge(i, j, **data)
        return G

    def to_adjacency_matrix(self, nodelist=None, sum_multiedges=True, weight="weight"):
        """
        Returns this Network as an adjacency matrix. A connection is represented by a int greater than zero, and zero
        means no connection is present.

        :param nodelist: defines the node order in the adjacency matrix; if no nodelist is provided, then the order
                         is determined by Network.nodes. (default is None)

        :param sum_multiedges : whether the presence of a connection should be indicated by summing the number of
                                edges in the connection (default is True)

        :param weight : a string or None indicating which edge attribute contains the edge weight. If an edge does
                        not contain the attribute, then 1 is used. (default is 'weight')

        :return: an adjacency matrix representing this Network
        """
        num_nodes = len(self._network)
        M = [[0]*num_nodes for _ in range(num_nodes)]
        if not nodelist:
            nodelist = list(self.nodes)
        for n1, n2, atts in self.edges:
            w = atts[weight] if weight in atts else 1
            i = nodelist.index(n1)
            j = nodelist.index(n2)
            if sum_multiedges:
                M[i][j] += w
            else:
                M[i][j] = w
        return M

    def _init_node(self, node_label):
        if node_label not in self._network:
            self._network[node_label] = self._new_node()

    def _new_node(self):
        return {"in": 0, "out": 0, "incoming": {}, "outgoing": []}

    def _increment_out_degree(self, node_label, n=1):
        self._network[node_label]["out"] += n

    def _decrement_out_degree(self, node_label, n=1):
        self._network[node_label]["out"] -= n

    def _increment_in_degree(self, node_label, n=1):
        self._network[node_label]["in"] += n

    def _decrement_in_degree(self, node_label, n=1):
        self._network[node_label]["in"] -= n

    def _add_outgoing(self, i, j):
        self._network[i]["outgoing"].append(j)

    def _remove_outgoing(self, i, j):
        self._network[i]["outgoing"] = [x for x in self._network[i]["outgoing"] if x != j]

    def _remove_incoming(self, i, j):
        if i in self._network[j]["incoming"]:
            del self._network[j]["incoming"][i]

    def __getitem__(self, item):
        return self._network[item]

    def __delitem__(self, key):
        self.remove_node(key)

    def __len__(self):
        return len(self._network)

    def __eq__(self, other):
        return self._network == other._network


class State:
    __slots__ = ("_activities", "_network", "_compression")

    def __init__(self, activities=None, network=None, compression=False):
        """

        :param activities: a dict from node label to activity

        :param network: a Network
        """
        self._compression = compression
        self._activities = None
        self.activities = activities
        self._network = None
        self.network = network

    @property
    def activities(self):
        if self._activities:
            return _Compressor.decompress(self._activities) if self._compression else self._activities

    @activities.setter
    def activities(self, a):
        if a:
            self._activities = _Compressor.compress(a) if self._compression else a

    @property
    def network(self):
        if self._network:
            return Network.decompress(self._network) if self._compression else self._network

    @network.setter
    def network(self, n):
        if n:
            self._network = n.compress() if self._compression else n


class _Compressor:
    def __init__(self):
        pass

    @staticmethod
    def compress(obj):
        """
        Compresses the given object into a binary format.

        :param network: an object

        :return: packed bytes
        """
        return msgpack.packb(obj)

    @staticmethod
    def decompress(packed):
        """
        Decompresses a compressed object.

        :param packed: packed bytes

        :return: a decompressed object
        """
        gc.disable()
        d = msgpack.unpackb(packed, strict_map_key=False)
        gc.enable()
        return d


import numpy as np
from .state import Network


class SubstitutionSystem:
    def __init__(self, axiom, rules, constants=None):
        self._rules = rules
        self._constants = constants if constants else []
        self._rules.update({c: c for c in self._constants})
        self._neighbourhood_size = self._get_neighbourhood_size(rules)
        self._initial_conditions = [s for s in axiom]
        self._dtype = type(self._initial_conditions[0])
        self._network = self._init_network(len(self._initial_conditions))
        self._last_node_index = len(self._initial_conditions) - 1
        self._num_nodes_added = 0
        self._last_timestep = 0

    def _get_neighbourhood_size(self, rules):
        neighbourhood_size = len(next(iter(rules)))
        for key in rules:
            if len(key) != neighbourhood_size:
                raise Exception("all keys in the rule map must be the same length")
        return neighbourhood_size

    def _init_network(self, n):
        # since we require Python 3.6, and dicts respect insertion order, we're using a plain dict here
        # (even though the 3.6 language spec doesn't officially support it)
        network = Network()
        for i in range(n):
            network.add_edge(i, i)
            for j in range(1, self._neighbourhood_size):
                if (i + j) < n:
                    network.add_edge(i + j, i)
        return network

    def activity_rule(self, ctx):
        ctx.remove_node(ctx.node_label)

        self._update_last_timestep(ctx)

        if len(ctx.neighbour_labels) != self._neighbourhood_size:
            return

        current_state = self._get_state(ctx)
        new_state = self._rules[current_state]

        for i, state in enumerate(new_state):
            self._last_node_index += 1
            new_node_label = self._last_node_index
            self._num_nodes_added += 1
            outgoing_links = {}
            for j in range(0, min(self._num_nodes_added, self._neighbourhood_size)):
                outgoing_links[new_node_label - j] = {}
            state = self._dtype(state)
            ctx.add_node(state, outgoing_links, new_node_label)

        # we return None here, since the graph has been re-written through the context, and the newly added node(s)
        #  carry the state information

    def _update_last_timestep(self, ctx):
        if self._last_timestep != ctx.timestep:
            self._num_nodes_added = 0
        self._last_timestep = ctx.timestep

    def _get_state(self, ctx):
        state = ""
        for neighbour in ctx.neighbour_labels:
            state += str(ctx.activity_of(neighbour))
        return state

    @property
    def network(self):
        return self._network

    @property
    def initial_conditions(self):
        return self._initial_conditions

    @staticmethod
    def pad(trajectory):
        activities = {t: s.activities for t, s in enumerate(trajectory)}
        activities = [[v for e, v in sorted(activities[k].items())] for k in sorted(activities)]
        max_len = np.max([len(a) for a in activities])
        return np.asarray([np.pad(a, (0, max_len - len(a)), 'constant', constant_values=0) for a in activities])

    @staticmethod
    def to_string(trajectory):
        s = []
        for state in trajectory:
            s.append("".join([state.activities[k] for k in sorted(state.activities)]))
        return s






class ReversibleRule:
    """
    An automaton rule explicitly set up to be reversible.
    """
    def __init__(self, activity_rule):
        """
        Creates a reversible automata rule by taking into consideration the previous state of a
        node, by taking the XOR of the rule's normal output with the previous state to get the new state.
        Note that this class works only for automata where the state of a node is binary.
        :param activity_rule: the automata rule to be used
        """
        self._activity_rule = activity_rule

    def __call__(self, ctx):
        regular_result = self._activity_rule(ctx)
        new_result = regular_result ^ ctx.past_activity_of(ctx.node_label)
        return new_result



from . import adjacency
import networkx as nx
import numpy as np
import netomaton as ntm


def cellular_automaton(n=200, r=1, periodic=True):
    adjacency_matrix = adjacency.cellular_automaton(n, r, boundary="periodic")
    return from_adjacency_matrix(adjacency_matrix)


def cellular_automaton2d(rows, cols, r=1, neighbourhood='Moore', periodic=True):
    adjacency_matrix = adjacency.cellular_automaton2d(rows, cols, r, neighbourhood, "periodic")
    return from_adjacency_matrix(adjacency_matrix)


def lattice(dim, periodic=False, self_loops=False, first_label=0):
    """
    Returns a bi-directional n-dimensional lattice (i.e. Euclidean) network.

    :param dim: a triple, representing the number of dimensions of the lattice

    :param periodic: whether the lattice is periodic (default is False)

    :param self_loops: if True, each node has a connection to itself (default is False)

    :param first_label: an integer specifying the first node label (default is 1)

    :return: a network
    """
    G = nx.grid_graph(dim=dim, periodic=periodic)
    G = nx.convert_node_labels_to_integers(G, first_label=first_label)
    network = ntm.Network()
    for edge in G.edges:
        if edge[0] == edge[1] and not self_loops:
            continue
        network.add_edge(edge[1], edge[0])
        network.add_edge(edge[0], edge[1])
    return network


def disconnected(nodes):
    """
    Create a fully disconnected network.

    :param nodes: either an int, containing the number of nodes, or a list, containing the node labels to use

    :return: a fully disconnected network
    """
    network = ntm.Network(n=nodes) if isinstance(nodes, int) else ntm.Network()
    if isinstance(nodes, (list, np.ndarray)):
        [network.add_node(n) for n in nodes]
    return network


def from_adjacency_matrix(adj):
    network = ntm.Network()
    for i, row in enumerate(adj):
        for j, val in enumerate(row):
            if val != 0:
                network.add_edge(i, j, weight=val)
    return network


def watts_strogatz_graph(n, k, p, seed=None):
    """
    Returns a Watts-Strogatz small-world graph as a Network.

    :param n: the number of nodes

    :param k: the number of nearest neighbours a node joins to

    :param p: the probability of re-connecting each edge

    :param seed: integer, random_state, or None (default); a random seed to use for random number generation

    :return: as Network
    """
    G = nx.watts_strogatz_graph(n, k, p, seed)
    adjacency = nx.adjacency_matrix(G).todense().tolist()
    return from_adjacency_matrix(adjacency)



from math import *

import matplotlib as mpl
import matplotlib.animation as animation
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from .utils import get_activities_over_time_as_list
from matplotlib.patches import RegularPolygon
from matplotlib.collections import PatchCollection


def _oddr_offset_to_pixel(x, y, size=1):
    p_x = size * sqrt(3) * (x + 0.5 * (y & 1))
    p_y = size * 3 / 2 * y
    return p_x, p_y


def _evenr_offset_to_pixel(x, y, size=1):
    p_x = size * sqrt(3) * (x - 0.5 * (y & 1))
    p_y = size * 3 / 2 * y
    return p_x, p_y


def _get_scaled_colormap(triples, colormap, vmin, vmax):
    vals = [t[2] for t in triples]
    vmin = min(vals) if vmin is None else vmin
    vmax = max(vals) if vmax is None else vmax
    cmap = plt.get_cmap(colormap)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    return cm.ScalarMappable(norm=norm, cmap=cmap)


def _get_triples(activities):
    # using the “odd-r” horizontal layout offset coordinate system described in:
    #   https://www.redblobgames.com/grids/hexagons/#coordinates
    # here we get a list of tuples with three values: (x, y, node value)
    return [(x, y, val) for y, row in enumerate(activities) for x, val in enumerate(row)]


def plot_hex_grid(trajectory, shape=None, slice=-1, title='', colormap='Greys', vmin=None, vmax=None, edgecolor=None):
    activities = get_activities_over_time_as_list(trajectory)
    if shape is not None:
        activities = np.array(activities).reshape((len(activities), shape[0], shape[1]))[slice]
    triples = _get_triples(activities)
    has_odd_rows = (len(activities) % 2) != 0
    fig, ax = plt.subplots(1)
    ax.set_aspect('equal')
    m = _get_scaled_colormap(triples, colormap, vmin, vmax)
    for t in triples:
        x, y = _oddr_offset_to_pixel(t[0], t[1]) if has_odd_rows else _evenr_offset_to_pixel(t[0], t[1])
        hex = RegularPolygon((x, y), numVertices=6, radius=1.,
                             orientation=np.radians(60),
                             facecolor=m.to_rgba(t[2]), edgecolor=edgecolor)
        ax.add_patch(hex)
    ax.scatter([t[0] for t in triples], [t[1] for t in triples], marker='')
    plt.gca().invert_yaxis()
    plt.title(title)
    plt.show()


def animate_hex(trajectory, title='', shape=None, save=False, interval=50, colormap='Greys',
                vmin=None, vmax=None, edgecolor=None):
    activities = get_activities_over_time_as_list(trajectory)
    if shape is not None:
        activities = np.reshape(activities, (len(activities), shape[0], shape[1]))
    triples = _get_triples(activities[0])
    has_odd_rows = (len(activities[0]) % 2) != 0
    fig, ax = plt.subplots(1)
    ax.set_aspect('equal')
    m = _get_scaled_colormap(triples, colormap, vmin, vmax)
    patches = []
    for t in triples:
        x, y = _oddr_offset_to_pixel(t[0], t[1]) if has_odd_rows else _evenr_offset_to_pixel(t[0], t[1])
        hex = RegularPolygon((x, y), numVertices=6, radius=1.,
                             orientation=np.radians(60), edgecolor=edgecolor)
        patches.append(hex)

    p = PatchCollection(patches, match_original=True, cmap=m.get_cmap())
    ax.add_collection(p)

    i = {'index': 0}
    def update(*args):
        i['index'] += 1
        if i['index'] == len(activities):
            i['index'] = 0
        new_triples = _get_triples(activities[i['index']])
        p.set_array(np.array([tr[2] for tr in new_triples]))
        return p,

    ax.scatter([t[0] for t in triples], [t[1] for t in triples], marker='')
    plt.gca().invert_yaxis()
    plt.title(title)

    ani = animation.FuncAnimation(fig, update, interval=interval, blit=True, save_count=len(activities))
    if save:
        ani.save('evolved.gif', dpi=80, writer="imagemagick")

    plt.show()




import os
import webbrowser
import networkx as nx


def configuration_to_nx(config):
    G = nx.MultiDiGraph()
    for relation in config:
        for i, node in enumerate(relation):
            G.add_node(node)
            if (i+1) < len(relation):
                from_node = node
                to_node = relation[i + 1]
                G.add_edge(from_node, to_node)
    return G


def show_network(G, name="out.html", is_notebook=False):
    html = get_html(G)
    o = open(name, "w")  # TODO write tempfile if no name is provided?
    o.write(html)
    o.close()
    if not is_notebook:
        webbrowser.open('file://' + os.path.realpath(name))
    # TODO if we are in a notebook, then there are some extra steps to render the html


def get_html(G):

    nodes = "[" + ",".join(["{id: " + str(n) + ", label: " + str(n) + ", shape: \"dot\", size: 10}" for n in G.nodes]) + "]"

    edges = []
    self_references = {}
    for e in G.edges:
        self_reference = ""
        if e[0] == e[1]:
            if (e[0], e[1]) not in self_references:
                self_references[(e[0], e[1])] = 0
            self_references[(e[0], e[1])] += 1
            self_reference = ", selfReferenceSize: %s" % (15 + self_references[(e[0], e[1])]*5)
        edges.append("{from: %s, to: %s, arrows: \"to\", weight: 1%s}" % (str(e[0]), str(e[1]), self_reference))
    edges = "[" + ",".join(edges) + "]"

    ret = """<html><head>
      <title>Netomaton</title>
      <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.19.1/vis.js"></script>
      <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.19.1/vis.css" rel="stylesheet" type="text/css">
      <style type="text/css">
          #mynetwork {
            width: 100%;
            height: 100%;
            border: 1px solid lightgray;
          }
          #loadingBar {
            position: absolute;
            top: 50%;
            left: 50%;
            -moz-transform: translateX(-50%) translateY(-50%);
            -webkit-transform: translateX(-50%) translateY(-50%);
            transform: translateX(-50%) translateY(-50%);
            -webkit-transition: all 0.5s ease;
            -moz-transition: all 0.5s ease;
            -ms-transition: all 0.5s ease;
            -o-transition: all 0.5s ease;
            transition: all 0.5s ease;
            opacity: 1;
          }
          #wrapper {
            position: relative;
            width: 100%;
            height: 100%;
          }
    
          #text {
            position: absolute;
            top: 8px;
            left: 330px;
            width: 30px;
            height: 50px;
            margin: auto auto auto auto;
            font-size: 16px;
            font-family: sans-serif;
            color: #000000;
          }
    
          div.outerBorder {
            position: relative;
            top: 0px;
            width: 400px;
            height: 44px;
            margin: auto auto auto auto;
          }
    
          #border {
            position: absolute;
            top: 10px;
            left: 10px;
            width: 300px;
            height: 13px;
            margin: auto auto auto auto;
            box-shadow: 0px 0px 4px rgba(0, 0, 0, 0.2);
          }
    
          #bar {
            position: absolute;
            top: 0px;
            left: 0px;
            width: 10px;
            height: 10px;
            margin: auto auto auto auto;
            border: 2px solid rgba(30, 30, 30, 0.05);
            background: rgb(0, 173, 246); /* Old browsers */
            box-shadow: 2px 0px 4px rgba(0, 0, 0, 0.4);
          }
      </style>
    <script type="text/javascript">
    function draw() {
      var nodes = new vis.DataSet(""" + nodes + """);
      var edges = new vis.DataSet(""" + edges + """);
      var container = document.getElementById('mynetwork');
      var data = {
        nodes: nodes,
        edges: edges
      };
      var options = {
          "physics": {
            "forceAtlas2Based": {
              "gravitationalConstant": -50,
              "centralGravity": 0.01,
              "springLength": 100,
              "springConstant": 0.53,
              "avoidOverlap": 0
            },
            "minVelocity": 0.75,
            "solver": "forceAtlas2Based",
            "timestep": 0.22
          }
      };
      var network = new vis.Network(container, data, options);
      
      network.on("stabilizationProgress", function (params) {
        var maxWidth = 296;
        var minWidth = 12;
        var widthFactor = params.iterations / params.total;
        var width = Math.max(minWidth, maxWidth * widthFactor);
    
        document.getElementById("bar").style.width = width + "px";
        document.getElementById("text").innerText =
          Math.round(widthFactor * 100) + "%";
      });
      network.once("stabilizationIterationsDone", function () {
        document.getElementById("text").innerText = "100%";
        document.getElementById("bar").style.width = "296px";
        document.getElementById("loadingBar").style.opacity = 0;
        // really clean the dom element
        setTimeout(function () {
          document.getElementById("loadingBar").style.display = "none";
        }, 500);
      });
    }
    </script>
    </head>
    <body onload="draw()">
    <div id="wrapper">
      <div id="mynetwork"></div>
      <div id="loadingBar">
        <div class="outerBorder">
          <div id="text">0%</div>
          <div id="border">
            <div id="bar"></div>
          </div>
        </div>
      </div>
    </div>
    </body></html>"""
    return ret


import numpy as np

from .topology import cellular_automaton2d


class Sandpile:
    def __init__(self, rows, cols, is_closed_boundary=True):
        self._K = 4  # this value is hard-coded because the neighbourhood type, "von Neumann", is fixed
        self._network = cellular_automaton2d(rows=rows, cols=cols, neighbourhood="von Neumann")
        self._boundary_indices = self._get_boundary_indices((rows, cols))
        self._is_closed_boundary = is_closed_boundary

    def _get_boundary_indices(self, shape):
        m = np.arange(shape[0]*shape[1]).reshape(shape)
        return np.concatenate((m[0], m[-1], m[:, 0], m[:, -1]), axis=None)

    def activity_rule(self, ctx):
        if self._is_closed_boundary and ctx.node_label in self._boundary_indices:
            return 0  # closed boundary conditions

        new_activity = ctx.current_activity

        neighbour_activities = list(ctx.neighbourhood_activities)
        neighbour_activities.pop(ctx.neighbour_labels.index(ctx.node_label))

        for neighbour_activity in neighbour_activities:
            if neighbour_activity >= self._K:
                new_activity += 1

        if ctx.current_activity >= self._K:
            new_activity -= self._K

        return new_activity

    @property
    def network(self):
        return self._network



import netomaton as ntm
import numpy as np


class FungalGrowthModel:
    """
    In this model, a network of agents are connected to eachother in a Euclidean lattice. Each agent
    possesses a certain amount of "resource". An agent's resource can change by picking up resource from
    a resource layer, and through the receipt of resource from its neighbours.

    An agent represents a cell, and the links between agents represent the flow of resources between cells. (The
    flow of resources is unidirectional only.)

    This implementation represents process "1", model "a", from the paper:
    Smith, David MD, et al. "Network automata: Coupling structure and function in dynamic networks."
    Advances in Complex Systems 14.03 (2011): 317-339.
    """
    def __init__(self, R_E, width, height, initial_conditions, resource_layer=None, seed=None, verbose=True):
        """
        :param R_E: resource absorption rate
        :param width: an integer representing the width of the grid of agents
        :param height: an integer representing the height of the grid of agents
        :param initial_conditions: a list of integers representing the initial conditions, in terms of the starting
                                   amount of resource for each agent
        :param resource_layer: a boolean matrix (flattened to a vector) representing the availability of resource
                               to an agent (default is None)
        :param seed: integer, random_state, or None (default); a random seed to use for random number generation
        :param verbose: whether to print progress statements to the console (default is True)
        """
        self._R_E = R_E
        self._d = 4  # the maximum node degree allowable
        if seed:
            np.random.seed(seed)
        self._verbose = verbose

        self._num_agents = width * height

        underlying_network = ntm.topology.lattice(dim=(1, width, height), periodic=True, first_label=0)
        self._links = set()
        for i, j, _ in underlying_network.edges:
            self._links.add(frozenset((i, j)))
        self._initial_network = ntm.Network(self._num_agents)

        # a boolean matrix (flattened to a vector) representing the availability of resource to an agent
        if resource_layer:
            assert len(resource_layer) == self._num_agents, \
                "the size of the resource layer must match the number of agents"
            self._resource_layer = resource_layer
        else:
            self._resource_layer = [0 for i in range(self._num_agents)]
            # provide a single infinite resource at the same location as the initial agent(s)
            for nz in np.nonzero(initial_conditions)[0]:
                self._resource_layer[nz] = 1

        if self._verbose:
            print("initialization complete")

    def topology_rule(self, ctx):
        if self._verbose:
            print("topology rule - t: %s" % ctx.timestep)
        curr_network = ctx.network
        new_network = curr_network.copy()
        for i, j in self._links:
            phi_S_i = 1 if ctx.activities[i] > 0 else 0
            phi_S_j = 1 if ctx.activities[j] > 0 else 0
            A_i_j = 1 if curr_network.has_edge(i, j) else 0
            A_j_i = 1 if curr_network.has_edge(j, i) else 0

            if A_i_j == 0 and A_j_i == 0 and phi_S_i == 0 and phi_S_j == 0:
                pass

            elif A_i_j == 0 and A_j_i == 0 and phi_S_i == 0 and phi_S_j == 1:
                p = 1 / (self._d - curr_network.degree(j))
                if np.random.random() < p:
                    new_network.add_edge(j, i)

            elif A_i_j == 0 and A_j_i == 0 and phi_S_i == 1 and phi_S_j == 0:
                p = 1 / (self._d - curr_network.degree(i))
                if np.random.random() < p:
                    new_network.add_edge(i, j)

            elif A_i_j == 0 and A_j_i == 0 and phi_S_i == 1 and phi_S_j == 1:
                p = 0.5
                if np.random.random() < p:
                    new_network.add_edge(i, j)
                else:
                    new_network.add_edge(j, i)

        for i, j, _ in new_network.edges:
            new_network.update_edge(i, j, weight=1/new_network.out_degree(i))

        return new_network

    def activity_rule(self, ctx):
        # S_i = sum of neighbourhood activities times corresponding weights + available resources from resource layer
        neighbour_contribution = 0. if len(ctx.neighbour_labels) > 0 else ctx.current_activity
        for neighbour in ctx.neighbour_labels:
            neighbour_contribution += ctx.connection_states[neighbour][0]["weight"] * ctx.activities[neighbour]

        phi = 1 if ctx.current_activity > 0 else 0  # whether this agent is alive or dead
        L = self._resource_layer[ctx.node_label]  # whether resource is available to the agent from the resource layer
        resource_layer_contribution = self._R_E * phi * L

        S_i = neighbour_contribution + resource_layer_contribution
        return S_i

    @property
    def network(self):
        return self._initial_network

    @property
    def update_order(self):
        return ntm.UpdateOrder.TOPOLOGY_FIRST

    @property
    def copy_network(self):
        return False



import numpy as np
from statistics import mode, StatisticsError
from collections import deque


def totalistic_ca(k, rule):
    """
    The totalistic rule as described in NKS. The average color is mapped to a whole number in [0, k - 1].
    The rule number is in base 10, but interpreted in base k. For a 1-dimensional cellular automaton, there are
    3k - 2 possible average colors in the 3-cell neighbourhood. There are n(k - 1) + 1 possible average colors for a
    k-color cellular automaton with an n-cell neighbourhood.

    :param k: the number of colors in this cellular automaton, where only 2 <= k <= 36 is supported

    :param rule: the k-color cellular automaton rule number in base 10, interpreted in base k

    :return: a function that represents the corresponding cellular automaton rule, that returns the result, a number
             from 0 to k - 1, of applying the given rule on the given state
    """
    # e.g. np.base_repr(777, base=3) -> '1001210'; the zfill pads the string with zeroes: '1'.zfill(3) -> '001'
    #   Bases greater than 36 not handled in base_repr.

    def _rule(ctx):
        neighbourhood = np.array(ctx.neighbourhood_activities)
        n = neighbourhood.size
        rule_string = np.base_repr(rule, base=k).zfill(n*(k - 1) + 1)
        if len(rule_string) > n*(k - 1) + 1:
            raise Exception("rule number out of range")
        neighbourhood_sum = np.sum(neighbourhood)
        # the rightmost element of the rule is for the average color 0, in NKS convention
        return int(rule_string[n*(k - 1) - neighbourhood_sum], k)
    return _rule


def majority_rule(ctx):
    """
    Returns the value of the most frequent state in the neighbourhood. If all states are equally frequent, then a
    random state is chosen from the neighbourhood.

    :param ctx: the NodeContext for a node

    :return: the new activity for the node with the given neighbourhood
    """
    try:
        return mode(ctx.neighbourhood_activities)
    except StatisticsError:
        # a StatisticsError is raised if there is a tie, and there is no single most common element
        return np.random.choice(ctx.neighbourhood_activities)


def _bits_to_int(bits):
    total = 0
    for shift, j in enumerate(bits[::-1]):
        if j:
            total += 1 << shift
    return total


def _int_to_bits(num, num_digits):
    converted = list(map(int, bin(num)[2:]))
    return np.pad(converted, (num_digits - len(converted), 0), 'constant')


def shift_to_center(activities, node_indices, node_index):
    center = len(activities) // 2
    shifted = deque(activities)

    def index_of(arr, val):
        return np.where(arr == val)[0][0] if type(arr) == np.ndarray else arr.index(val)
    shifted.rotate(center - index_of(node_indices, node_index))
    return list(shifted)


def binary_ca_rule(rule, scheme=None):
    """
    Converts the given rule number to a binary representation, and uses this to determine the value to return.
    The process is approximately described as:
    1. convert state to int, so [1,0,1] -> 5, call this state_int
    2. convert rule to binary, so 254 -> [1,1,1,1,1,1,1,0], call this rule_bin_array
    3. new value is rule_bin_array[7 - state_int]
         we subtract 7 from state_int to be consistent with the numbering scheme used in NKS
         in NKS, rule 254 for a 1D binary cellular automaton is described as:
        [1,1,1]  [1,1,0]  [1,0,1]  [1,0,0]  [0,1,1]  [0,1,0]  [0,0,1]  [0,0,0]
           1        1        1        1        1        1        1        0
    If None is provided for the scheme parameter, the neighbourhoods are listed in lexicographic order (the reverse of
    the NKS convention). If 'nks' is provided for the scheme parameter, the NKS convention is used for listing the
    neighbourhoods.

    :param rule: an int indicating the cellular automaton rule number

    :param scheme: can be None (default) or 'nks'; if 'nks' is given, the rule numbering scheme used in NKS is used

    :return: a function that represents the corresponding cellular automaton rule, where the context activities are a
             binary array of length 2r + 1, and the function returns a result, 0 or 1, of applying the given rule on
             the given state
    """
    # shift the activities so that the current cell's activity is in the center

    def _rule(ctx):
        activities = shift_to_center(ctx.neighbourhood_activities, ctx.neighbour_labels, ctx.node_label)
        state_int = _bits_to_int(activities)
        n = 2**len(activities)
        rule_bin_array = _int_to_bits(rule, n)
        if scheme == 'nks':
            return rule_bin_array[(n-1) - state_int]
        return rule_bin_array[state_int]
    return _rule


def nks_ca_rule(rule):
    """
    A convenience function, that calls binary_rule with scheme = 'nks'.

    :param rule: an int indicating the cellular automaton rule number

    :return: a function that represents the corresponding cellular automaton rule, where the activities are a binary
             array of length 2r + 1
    """
    return binary_ca_rule(rule, scheme='nks')


def game_of_life_rule(ctx):
    activities = ctx.neighbourhood_activities
    center_cell = activities[len(activities) // 2]
    total = np.sum(activities)
    if center_cell == 1:
        if total - 1 < 2:
            return 0  # Any live cell with fewer than two live neighbours dies, as if caused by underpopulation.
        if total - 1 == 2 or total - 1 == 3:
            return 1  # Any live cell with two or three live neighbours lives on to the next generation.
        if total - 1 > 3:
            return 0  # Any live cell with more than three live neighbours dies, as if by overpopulation.
    else:
        if total == 3:
            return 1  # Any dead cell with exactly three live neighbours becomes a live cell, as if by reproduction.
        else:
            return 0


def wireworld_rule(ctx):
    if ctx.current_activity == 0:  # empty
        return 0
    if ctx.current_activity == 1:  # electron head
        return 2
    if ctx.current_activity == 2:  # electron tail
        return 3
    if ctx.current_activity == 3:  # conductor
        electron_head_count = ctx.neighbourhood_activities.count(1)
        return 1 if electron_head_count == 1 or electron_head_count == 2 else 3



from .asynchronous_rule import *
from .topology import from_adjacency_matrix
import numpy as np


class HopfieldNet:
    """
    The Hopfield net implemented here expects bipolar ({-1,1}), and not binary ({0,1}), values. It consists of a
    training step, where the weights of the adjacency matrix are established, using a simple Hebbian learning rule,
    and an evaluation step, in which a starting pattern is evolved for a specified number of timesteps. The final state
    of the network should settle into one of the training patterns.
    """
    def __init__(self, n):
        self._activity_rule = AsynchronousRule(activity_rule=self._rule, n=n, randomize_each_cycle=True)
        self._num_nodes = n

    def train(self, P):
        """
        The training set consists of patterns to be learned by this net. The patterns should be composed of
        bipolar ({-1,1}), and not binary ({0,1}), values. The learned weights will be stored in, and comprise, the
        adjacency matrix. Therefore, before the network can be used to evaluate a pattern, it must be trained.
        :param P: the set of training patterns, as a list of lists
        """
        self._adjacency_matrix = np.zeros((len(P[0]), len(P[0])), dtype=np.int)
        for p in P:
            for i in range(len(p)):
                for j in range(len(p)):
                    if i ==j:
                        self._adjacency_matrix[i, j] = 0
                    else:
                        self._adjacency_matrix[i, j] += p[i]*p[j]
        self._network = from_adjacency_matrix(self._adjacency_matrix)

    def _rule(self, ctx):
        """
        Peforms a linear combination of the neighbourhood activities and the corresponding weights. If the combined
        value is greater than or equal to 0, then 1 is returned, otherwise -1 is returned.
        :param ctx: the NodeContext; note that the neighbourhood of the node, in this case, does not include the
                    node itself
        :return: the new value of the node
        """
        V = 0
        for neighbour_label in ctx.neighbour_labels:
            V += ctx.connection_states[neighbour_label][0]["weight"] * ctx.activities[neighbour_label]
        return 1 if V >= 0 else -1

    @property
    def activity_rule(self):
        return self._activity_rule

    @property
    def network(self):
        return self._network

    @property
    def adjacency_matrix(self):
        return self._adjacency_matrix

    @property
    def num_nodes(self):
        return self._num_nodes



import math
from .utils import get_activities_over_time_as_list
from .state import State
import numpy as np


def shannon_entropy(string):
    """
    Calculates the Shannon entropy for the given string.

    :param string: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the Shannon entropy
    """
    symbols = dict.fromkeys(list(string))
    symbol_probabilities = [float(string.count(symbol)) / len(string) for symbol in symbols]
    H = -sum([p_symbol * math.log(p_symbol, 2.0) for p_symbol in symbol_probabilities])
    return H + 0  # add 0 as a workaround so we don't end up with -0.0


def average_node_entropy(trajectory_or_activities):
    """
    Calculates the average node entropy in the given automaton evolution, where entropy is the Shannon entropy.
    The state of a node over time is represented as a string, and its entropy is calculated. The same is done for all
    nodes in this Network Automaton, and the average entropy is returned.

    :param trajectory_or_activities: a list of States that represents the evolution of the automaton, or a list of
                                     lists that represents the evolution of the automaton, where an inner list
                                     represents the activities of the nodes at a given timestep

    :return: a real number representing the average node Shannon entropy
    """
    if len(trajectory_or_activities) is 0:
        raise Exception("there are no activities")
    if isinstance(trajectory_or_activities[0], State):
        # trajectory_or_activities is a trajectory, convert it to an activities list
        activities = get_activities_over_time_as_list(trajectory_or_activities)
    else:
        activities = trajectory_or_activities
    num_cols = len(activities[0])
    entropies = []
    for i in range(0, num_cols):
        node_states_over_time = ''.join([str(x) for x in [y[i] for y in activities]])
        entropy = shannon_entropy(node_states_over_time)
        entropies.append(entropy)
    return np.mean(entropies)


def joint_shannon_entropy(stringX, stringY):
    """
    Calculates the joint Shannon entropy between the given strings, which must be of the same length.

    :param stringX: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :param stringY: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the joint Shannon entropy between the given strings
    """
    X = np.array(list(stringX))
    Y = np.array(list(stringY))
    joint_symbol_probabilities = []
    for x in set(X):
        for y in set(Y):
            joint_symbol_probabilities.append(np.mean(np.logical_and(X == x, Y == y)))
    return sum(-p * np.log2(p) for p in joint_symbol_probabilities if p != 0)


def mutual_information(stringX, stringY):
    """
    Calculates the mutual information between the given strings, which must be of the same length.

    :param stringX: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :param stringY: any string, such as '000101001', '12402', or 'aBcd1234ef5g'

    :return: a real number representing the mutual information between the given strings
    """
    return shannon_entropy(stringX) + shannon_entropy(stringY) - joint_shannon_entropy(stringX, stringY)


def average_mutual_information(trajectory_or_activities, temporal_distance=1):
    """
    Calculates the average mutual information between a node and itself at the next n time steps, given by the
    specified temporal distance. A temporal distance of 1 means the next time step.
    For example, consider the following string, '00101010110', which represents the state of a node over 11 time steps.
    The strings which will be used for the computation of the mutual information between a node and itself at the
    next time step are: '0010101011' and '0101010110', since we pair each time-step value with its next value:
    " 00101010110"
    "00101010110 "

    :param trajectory_or_activities: a list of States that represents the evolution of the automaton, or a list of
                                     lists that represents the evolution of the automaton, where an inner list
                                     represents the activities of the nodes at a given timestep

    :param temporal_distance: the size of temporal separation, where the value must be greater than 0 and
                              less than the number of time steps.

    :return: a real number representing the average mutual information between a node and itself at the next time step
    """
    if len(trajectory_or_activities) is 0:
        raise Exception("there are no activities")
    if isinstance(trajectory_or_activities[0], State):
        # trajectory_or_activities is a trajectory, convert it to an activities list
        activities = get_activities_over_time_as_list(trajectory_or_activities)
    else:
        activities = trajectory_or_activities
    num_cols = len(activities[0])
    if not (0 < temporal_distance < num_cols):
        raise Exception("the temporal distance must be greater than 0 and less than the number of time steps")
    mutual_informations = []
    for i in range(0, num_cols):
        node_states_over_time = ''.join([str(x) for x in [y[i] for y in activities]])
        mi = mutual_information(node_states_over_time[:-temporal_distance], node_states_over_time[temporal_distance:])
        mutual_informations.append(mi)
    return np.mean(mutual_informations)



from abc import abstractmethod

from .topology import cellular_automaton, from_adjacency_matrix
from .utils import get_activities_over_time_as_list


class TuringMachine:
    LEFT = 0
    STAY = 1
    RIGHT = 2

    @abstractmethod
    def network(self):
        pass

    @abstractmethod
    def activity_rule(self, ctx):
        pass


class TapeCentricTuringMachine(TuringMachine):
    """
    A Turing Machine modelled as a Network Automaton with a number of nodes representing the tape (with the same local
    connectivity as an Elementary Cellular Automaton), whose states are mutated as the tape is written to, and separate
    variables for the state and current location of the head.
    """

    def __init__(self, n, rule_table, initial_head_state, initial_head_position):
        self._n = n
        self._rule_table = rule_table
        self._head_history = [(initial_head_state, initial_head_position)]
        self._current_timestep = 1

    @property
    def network(self):
        return cellular_automaton(self._n)

    def activity_rule(self, ctx):
        if ctx.timestep != self._current_timestep:
            self._current_timestep = ctx.timestep
        head_state, head_pos = self._head_history[self._current_timestep - 1]
        node_state = ctx.current_activity
        if ctx.node_label == head_pos:
            try:
                next_head_state, new_node_state, direction = self._rule_table[head_state][node_state]
            except KeyError as err:
                raise Exception("no rule defined for head state %s, input %s" % (head_state, node_state)) from err
            self._head_history.append((next_head_state, ctx.neighbour_labels[direction]))
            return new_node_state
        return node_state

    def head_activities(self, trajectory):
        activities = get_activities_over_time_as_list(trajectory)
        annotations = [[None for _ in range(self._n)] for _ in range(len(activities))]
        for i, h in enumerate(self._head_history):
            annotations[i][h[1]] = str(h[0])
        return annotations


class HeadCentricTuringMachine(TuringMachine):
    """
    A Turing Machine modelled as a Network Automaton with a single node that carries the state of the head, and a
    separate tape that is read from and written to during processing.
    """

    def __init__(self, tape, rule_table, initial_head_state, initial_head_position,
                 terminating_state=None, max_timesteps=None):
        if terminating_state is None and max_timesteps is None:
            raise Exception("a terminating state or the max number of timesteps must be specified")
        self._tape_history = [tape]
        self._rule_table = rule_table
        # the state of the single node in the Network Automaton is a tuple: (head_state, head_position)
        self._initial_conditions = [(initial_head_state, initial_head_position)]
        self._head_pos = initial_head_position
        self._terminating_state = terminating_state
        self._max_timesteps = max_timesteps
        self._halt = False

    @property
    def network(self):
        # a Turing Machine can be thought of as a Network Automaton with a single node
        return from_adjacency_matrix([[1]])

    @property
    def initial_conditions(self):
        return self._initial_conditions

    def activity_rule(self, ctx):
        input_from_tape = ctx.input
        head_state, head_pos = ctx.current_activity
        try:
            next_head_state, val_to_write, direction = self._rule_table[head_state][input_from_tape]
        except KeyError as err:
            raise Exception("no rule defined for head state %s, input %s" % (head_state, input_from_tape)) from err
        new_tape = [i for i in self._tape_history[-1]]
        new_tape[head_pos] = val_to_write
        self._tape_history.append(new_tape)

        if self._terminating_state is not None and next_head_state == self._terminating_state:
            self._halt = True

        self._head_pos = self._next_pos(direction, head_pos)
        return next_head_state, self._head_pos

    def _next_pos(self, direction, head_pos):
        if direction == TuringMachine.LEFT:
            return (head_pos - 1) % len(self._tape_history[-1])
        elif direction == TuringMachine.RIGHT:
            return (head_pos + 1) % len(self._tape_history[-1])
        elif direction == TuringMachine.STAY:
            return head_pos
        else:
            raise Exception("unsupported direction: %s" % direction)

    def input_function(self, t, a, n):
        if self._max_timesteps is not None and len(self._tape_history) == self._max_timesteps:
            return None

        if self._halt:
            return None

        return self._tape_history[-1][self._head_pos]

    def activities_for_plotting(self, trajectory):
        activities = get_activities_over_time_as_list(trajectory)
        head_activities = [[None for _ in range(len(self._tape_history[-1]))] for _ in range(len(self._tape_history))]
        for i, h in enumerate(activities):
            head_activities[i][h[0][1]] = str(h[0][0])

        return self._tape_history, head_activities



from setuptools import setup, find_packages

packages = find_packages(exclude=("tests", "demos", "doc", "resources",))

setup(name="netomaton",
      version="1.1.0",
      description="Netomaton, A Python library for working with Network Automata.",
      long_description="Netomaton is a Python framework for exploring discrete dynamical network systems, "
                       "also known as Network Automata. It is a software abstraction meant to aid in the "
                       "implementation of models of collective computation",
      license="Apache License 2.0",
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
      ],
      url='http://github.com/lantunes/netomaton',
      author="Luis M. Antunes",
      author_email="lantunes@gmail.com",
      packages=packages,
      keywords=["network automata", "cellular automata", "complexity", "complex systems", "computation", "non-linear dynamics"],
      python_requires='>3.5.2',
      install_requires=["numpy >= 1.15.4", "matplotlib >= 3.0.2", "networkx == 2.2", "scipy == 1.3.1", "msgpack == 1.0.2"])



import math
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np


class HopfieldTankTSPNet:
    """
    Based on J. J. Hopfield and D. W. Tank, "'Neural' Computation of Decisions in Optimization Problems",
    Biol. Cybern: 52, 141-152 (1985).
    """
    def __init__(self, points, dt=1.00, A=500, B=500, C=200, D=500, u_0=0.02, n=15):
        """
        A point is a tuple (x, y), where x is the x-coordinate and y is the y-coordinate of the point.
        :param points: a list of tuples, where each tuple represents a point's x and y coordinates
        :param dt: the size of the timestep
        :param A: a hyparameter
        :param B: a hyparameter
        :param C: a hyparameter
        :param D: a hyparameter
        :param u_0: a hyparameter
        :param n: a hyparameter
        """
        self._points = points
        self._dt = dt
        self._A = A
        self._B = B
        self._C = C
        self._D = D
        self._u_0 = u_0
        self._n = n
        self._distances_map = self._get_distances_map(self.get_distances())
        self._node_label_map = self._get_node_label_map(points)
        self._coordinate_map = self._get_coordinate_map(self._node_label_map)
        self._adjacency_matrix = self._get_adjacency_matrix(self._node_label_map)

    def get_distances(self):
        """
        A point is a tuple (x, y), where x is the x-coordinate and y is the y-coordinate of the point.
        :param points: a list of tuples, where each tuple represents a point's x and y coordinates
        :return: a list of n-choose-2 triples (A, B, d), where A is the index of point A, B is the index of point B,
                 and d is the distance between them
        """
        distances = []
        i = 0
        while i < len(self._points) - 1:
            from_point = self._points[i]
            for j in range(i + 1, len(self._points)):
                to_point = self._points[j]
                d = math.sqrt((to_point[0] - from_point[0]) ** 2 + (to_point[1] - from_point[1]) ** 2)
                distances.append((i, j, d))
            i += 1

        return distances

    def get_permutation_matrix(self, activities):
        if not isinstance(activities, np.ndarray):
            activities = np.array(activities)
        f = lambda u: self._V(u)
        return f(activities[-1].reshape(len(self._points), len(self._points)))

    def get_tour_graph(self, points, permutation_matrix):
        """
        A point is a tuple (x, y), where x is the x-coordinate and y is the y-coordinate of the point.
        The permutation_matrix is n x n matrix, where n is the number of points in the TSP. Each row represents a point,
        and each column represents the position of that point in the tour.
        :param points: the list of points (tuples) which represent the points in the TSP
        :param permutation_matrix: an n x n matrix, where n is the number of points, which represents the final tour
        :return: a NetworkX Graph representing the tour, a dictionary defining the NetworkX positions for the Graph, and
                 the total tour length
        """
        assert len(points) == len(permutation_matrix), \
            "the number of rows in permutation_matrix does not match the number of points"
        assert len(points) == len(permutation_matrix[0]), \
            "the number of columns in permutation_matrix does not match the number of points"

        G = nx.Graph()
        for point_index in range(len(points)):
            G.add_node(point_index)

        tour_index_to_point_index = {}

        for point_index, row in enumerate(permutation_matrix):
            tour_index = np.argmax(row)
            if tour_index in tour_index_to_point_index:
                raise Exception("a point has already claimed position #%s in the tour" % str(tour_index + 1))
            tour_index_to_point_index[tour_index] = point_index

        distances = []

        for tour_index, point_index in tour_index_to_point_index.items():
            from_point_index = point_index
            to_point_index = tour_index_to_point_index[(tour_index + 1) % len(points)]
            G.add_edge(from_point_index, to_point_index)

            from_point = points[from_point_index]
            to_point = points[to_point_index]
            distances.append(math.sqrt((to_point[0] - from_point[0]) ** 2 + (to_point[1] - from_point[1]) ** 2))

        pos = {i: point for i, point in enumerate(points)}

        return G, pos, sum(distances)

    def plot_tour(self, G, pos):
        """
        Renders the points and tour.
        :param G: a NetworkX Graph representing the points and the tour
        :param pos: a dictionary defining the NetworkX positions for the Graph
        """
        nx.draw_networkx(G, pos)
        plt.show()

    def _get_distances_map(self, distances):
        """
        Given a list of n-choose-2 triples (A, B, d), where A is the index of point A, B is the index of point B,
        and d is the distance between them, returns a dictionary where the keys are a tuple, (A, B), and the values are the
        distances between A and B.
        :param distances: a list of n-choose-2 triples (A, B, d), where A is the index of point A, B is the index of point B,
                          and d is the distance between them
        :return: a dictionary where the keys are a tuple, (A, B), and the values are the
                 distances between A and B
        """
        distances_map = {}
        for a, b, d in distances:
            distances_map[(a, b)] = d
            distances_map[(b, a)] = d
            distances_map[(a, a)] = 0.0
            distances_map[(b, b)] = 0.0
        return distances_map

    def _get_node_label_map(self, points):
        """
        Returns a dictionary where the keys are the node indices (there are n^2 nodes, where n is the number of points),
        and the values are a tuple (row, col), representing the row index and column index of the node in the permutation
        matrix (the matrix describing the tour, where each row represents a point, and each column represents the position
        of that point in the tour.
        :param points: a list of tuples, where each tuple represents a point's x and y coordinates
        :return: a dictionary with node indices as keys and tuple (row, col) for the position of the node in the permutation
                 matrix as values
        """
        node_label_map = {}
        num_points = len(points)
        current_point = 0
        current_node = 0
        while current_point < num_points:
            for n in range(num_points):
                node_label_map[current_node] = (current_point, n)
                current_node += 1
            current_point += 1
        return node_label_map

    def _get_coordinate_map(self, node_label_map):
        coordinate_map = {}
        for node_index, coords in node_label_map.items():
            coordinate_map[coords] = node_index
        return coordinate_map

    def _get_adjacency_matrix(self, node_label_map):
        # fully connected network
        return [[1 for _ in range(len(node_label_map))] for _ in range(len(node_label_map))]

    @property
    def adjacency_matrix(self):
        return self._adjacency_matrix

    def _V(self, u):
        return (1/2) * (1 + np.tanh(u / self._u_0))

    def _get_opposite_neighbour_activity(self, activities, neighbour_indices, opposite_neighbour_label):
        for i, n in enumerate(neighbour_indices):
            if n == opposite_neighbour_label:
                return activities[i]
        raise Exception("could not find the opposite neighbour index: %d" % opposite_neighbour_label)

    def activity_rule(self, ctx):
        current_activity = ctx.current_activity
        node_row, node_col = self._node_label_map[ctx.node_label]

        A_sum = 0
        B_sum = 0
        C_sum = 0
        D_sum = 0
        for i, neighbour_activity in enumerate(ctx.neighbourhood_activities):
            neighbour_label = ctx.neighbour_labels[i]
            neighbour_node_row, neighbour_node_col = self._node_label_map[neighbour_label]

            if neighbour_node_row == node_row and neighbour_node_col != node_col:
                A_sum += self._V(neighbour_activity)

            if neighbour_node_col == node_col and neighbour_node_row != node_row:
                B_sum += self._V(neighbour_activity)

            # global inhibition
            C_sum += self._V(neighbour_activity)

            if neighbour_node_col == ((node_col - 1) % len(self._points)):
                opp_neighbour_label = self._coordinate_map[(neighbour_node_row, (node_col + 1) % len(self._points))]
                opp_neighbour_activity = self._get_opposite_neighbour_activity(ctx.neighbourhood_activities, ctx.neighbour_labels, opp_neighbour_label)
                D_sum += (self._distances_map[(neighbour_node_row, node_row)] * (self._V(neighbour_activity) + self._V(opp_neighbour_activity)))

        activity = (-current_activity) - (self._A * A_sum) - (self._B * B_sum) - (self._C * (C_sum - self._n)) - (self._D * D_sum)

        return current_activity + (self._dt * activity)

